(* ::Package:: *)

BeginPackage["BuTools`Own`"];


Unprotect[ones2, RationalToMarkovian2, METoPH2, RndTransf, PH2PHuniv2, MinMaxLag1Corr, MaxLag1Corr, MaxRangeLag1Corr, EntBuchholz, MAPSim, VectToBuchh, MatToBuchh, MomsFromSim, JMomsFromSim, MomsToBuchh,
JMomsToBuchh, BuchhToMat, BuchhToVect, BuchhToJmoms, PhFitToPH, MinimizeJmomDiff];


Begin["`Private`"]

Needs["BuTools`Utilities`"]
Needs["BuTools`PH`"]
Needs["BuTools`MAP`"]



(* Function for generating summing vector [1 1 ... 1] *) 
ones2[i_]:=Table[1,{i}];


(* Modified functions of METoPH;
returns 1 as res, if Markovian representation is found, otherwise returns 0;
TODO: check if needed *)
RationalToMarkovian2[Xin_, size_,TransFunc_,EvalFunc_,CheckFunc_] :=
Module[ {gsum,sum, n, X, Y,  nstepsize,nrepeat,niter,nerrorfunc,errortype,stepsize,i,j,ErrorFunc,x},
X=Xin;
gsum=0;
nstepsize=10;(*10;*) (* a-t csokkentem 2^-1 tol 2^-nstepsize-ig *)
nrepeat=1;(*2;*) (* minden a-ra n2 kort fut *)
niter=50; (*500;*) (* adott celfuggvenyre max niter iteraciot csinal *)
nerrorfunc=4; (* celfuggvenyre max n3 iteraciot csinal *)

stepsize=1.;
For[i=1,i<=nstepsize,i++,
stepsize=stepsize/2;
For[j=1,j<=nrepeat,j++,
sum=0;
For[errortype=1,errortype<=nerrorfunc,errortype++,
If[errortype==1,ErrorFunc[x_]:=E^(-100 x)];
If[errortype==2,ErrorFunc[x_]:=E^(-1000 x)];
If[errortype==3,ErrorFunc[x_]:=E^(-2 x)];
If[errortype==4,ErrorFunc[x_]:=E^-(x-1)^3];

{X,n}=ErrorMinimize[X,size,niter,stepsize,ErrorFunc,TransFunc,EvalFunc,CheckFunc];
If[CheckFunc[X],
Print["Found a Markovian representation in ", gsum+sum+n, " iterations !!!"];
Return[X~Join~{1}];
];
sum+=n;
];
];
gsum+=sum;
];
Print["No MARKOVIAN representation is found in ", gsum, " iterations !!!"];
Return[X~Join~{0}];
];


METoPH2[\[Alpha]_,A_]:= 
Module[{size},
size=Dimensions[A][[1]];
X={\[Alpha],A};
Return[RationalToMarkovian2[ X,size, PHTransFunc, PHEvalFunc, PHCheckFunc]];
];


PHCheckFunc[X_] :=CheckPH[X[[1]],X[[2]]];

PHTransFunc[X_,i_,j_,a_] :=
Module[ {bb,size},
size=Dimensions[X[[2]]][[1]];
bb=IdentityMatrix[size];
bb[[i,j]]=a;
bb[[i,i]]=1-a;
Return[{X[[1]].bb,Inverse[bb].X[[2]].bb}]
];
PHEvalFunc[X_,Error_] :=
Module[ {i,j,sum,size,\[Alpha],A,a},
size=Dimensions[X[[1]]][[1]];
\[Alpha]=X[[1]];
A=X[[2]];
a=-A.Table[1,{size}];
sum=0;
For[i=1,i<=size,i++,
sum+=Error[\[Alpha][[i]]  ];];
For[i=1,i<=size,i++,
For[j=1,j<=size,j++,
If[i!=j,sum+=Error[A[[i,j]]  ] ];];];
For[i=1,i<=size,i++,
sum+=Error[a[[i]]  ];];
Return[sum];
];



(* Transformation of an ME representarion to an other *)
RndTransf[\[Alpha]_,D0_]:=
	Module[{T,size},
		size=Length[\[Alpha]];
		T=RandomReal[{0,1},{size,size}];
		T/=T.Table[1,{size}];
		Return[{\[Alpha].T,Inverse[T].D0.T}]
	];


(* MOD: new parameter: mes, to choose the measure which we optimize for;
   MOD2: new optional parameter: jmoms, in case of trying to minimize joint moment based distance
   MOD3: final value of goal function to be returned too
*)
PH2PHuniv2[\[Alpha]in_, h0in_, mes_, jmoms_:{}] :=
Module[ {gsum,sum, n,\[Alpha], h0, increase, n1,n2,n3, size,cum, a, i, j,print, goalF},
size=Dimensions[h0in][[1]];
print=False;

If[!CheckPH[\[Alpha]in, h0in],Return[{0,0}]];


cum=IdentityMatrix[size];
\[Alpha]=\[Alpha]in; h0=h0in; gsum=0;

If[print,Print[" initial input vector=",\[Alpha] ," initial output vector=",PHout[\[Alpha],h0] ]];
a=Entrophy[Table[1/size,{size}]]//N;
If[print,Print["size=",size, " maximal entrophy for input=",a, " maximal entrophy for input and output=",2a]];
a=Entrophy[\[Alpha]];
If[print,Print[" initial entrophy for input=",a, " initial entrophy for output=",Entrophy[PHout[\[Alpha],h0]], " initial entrophy for input and output=",a+Entrophy[PHout[\[Alpha],h0]]]];


n1=10; (* a-t csokkentem 2^-1 tol 2^-n1-ig *)
n2=1; (* minden a-ra n2 kort fut *)
n3=500; (* celfuggvenyre max n3 iteraciot csinal *)

MeasureEnt[\[Alpha]_,h0_]:=Entrophy[\[Alpha]];
(* MOD: Maximization of the production of entropies seems better *)

(* Optimizing for maximal entropy product - default *)
MeasureEnt[\[Alpha]_,h0_]:=Entrophy[\[Alpha]]*Entrophy[PHout[\[Alpha],h0]];

(* Optimizing for maximal entropy sum *)
If[mes==0,MeasureEnt[\[Alpha]_,h0_]:=Entrophy[\[Alpha]]+Entrophy[PHout[\[Alpha],h0]]];

(* Optimizing for maximal entropy product *)
If[mes==1,MeasureEnt[\[Alpha]_,h0_]:=Entrophy[\[Alpha]]*Entrophy[PHout[\[Alpha],h0]]];

(* Optimizing for minimal lag1 correlation *)
If[mes==11,MeasureEnt[\[Alpha]_,h0_]:=-MinMaxLag1Corr[\[Alpha],h0][[1]]];

(* Optimizing for maximal lag1 correlation *)
If[mes==12,MeasureEnt[\[Alpha]_,h0_]:=MinMaxLag1Corr[\[Alpha],h0][[3]]];

(* Optimizing for maximal range of lag1 correlation *)
If[mes==13,MeasureEnt[\[Alpha]_,h0_]:=MinMaxLag1Corr[\[Alpha],h0][[3]]-MinMaxLag1Corr[\[Alpha],h0][[1]]];

(* MOD2: Optimizing for minimal distance in jmoms *)
If[mes==20,MeasureEnt[\[Alpha]_,h0_]:=-MinimizeJmomDiff[\[Alpha],h0,jmoms][[1]] ];

(* Optimizing for maximal out entropy *)
If[mes==-1,MeasureEnt[\[Alpha]_,h0_]:=Entrophy[PHout[\[Alpha],h0]]];

(* Optimizing for maximal in entropy *)
If[mes==-2,MeasureEnt[\[Alpha]_,h0_]:=Entrophy[\[Alpha]]];


a=1.;
For[i=1,i<=n1,i++,
a=a/2;
For[j=1,j<=n2,j++,
sum=0;
increase=True; n=0;
If[print,Print[" initial entrophy=",MeasureEnt[\[Alpha],h0]," a=",a]];
While[increase&&(n<n3),{\[Alpha], h0,cum,increase}=MeasureMax[\[Alpha],h0,cum,a];n+=1 ];
If[print,Print["entrophy=",MeasureEnt[\[Alpha],h0]," n=",n-1," a=",a," \[Alpha]=",\[Alpha],  " h0=",h0]];

sum+=n-1;

If[print,Print["Number of iterations:", sum]];
gsum+=sum;
];
];
(*
Print[" ---------------------- "];

Print["This is the best representation obtained in ", gsum, " iterations !!!"];
a=Entrophy[\[Alpha]];
Print[" final input vector=",\[Alpha] ," final output vector=",PHout[\[Alpha],h0] ];
Print[" final entrophy for input=",a, " final entrophy for output=",Entrophy[PHout[\[Alpha],h0]], " final entrophy for input and output=",a+Entrophy[PHout[\[Alpha],h0]]];
*)

(* MOD3: returning final value of goal function too *)
goalF=MeasureEnt[\[Alpha], h0];

Return[{\[Alpha],h0, goalF}];
]



(* TODO: clean the situation of these *)
(* Function for measuring entrophy *)
MeasureEnt[\[Alpha]_,h0_]:=Entrophy[\[Alpha]]+Entrophy[PHout[\[Alpha],h0]];

(* PH2PH functions *)

PHout[\[Alpha]_,h0_ ] :=
Module[ {v1, v2,size},
size=Dimensions[h0][[1]];
v1=\[Alpha].Inverse[-h0];
v2= - Table[1,{size}].Transpose[h0];
Return[v1 v2]
]

Entrophy[v_] :=
Module[ {i,j,size,sum},
size=Dimensions[v][[1]];
sum=0;
For[i=1,i<=size,i++,
If[v[[i]]>0,
sum=sum -v[[i]] Log[v[[i]]]];
];
Return[sum]
]

PHTrafo[\[Alpha]_,h0_,cum_,i_,j_,a_] :=
Module[ {bb,size},
size=Dimensions[h0][[1]];
bb=IdentityMatrix[size];
bb[[i,j]]=a;
bb[[i,i]]=1-a;
Return[{\[Alpha].bb,Inverse[bb].h0.bb,cum.bb}]
]

CheckPHEnt[\[Alpha]_,h0_] :=
Module[ {k,myh0,size,hh},
size=Dimensions[h0][[1]];
hh=Table[1,{size}];
myh0=h0;
For[k=1,k<=size,k++,myh0[[k,k]]=-h0[[k]].hh];
If[ Min[\[Alpha]]>=0 && Min[myh0]>=0,
Return[True], Return[False]];
]

MeasureMax[\[Alpha]_,h0_,cum_,a_] :=
Module[ {i,j,k,size,m, mbe,h0u,\[Alpha]u,cumu,max,maxi,maxj,maxa},
mbe=MeasureEnt[\[Alpha],h0];
max=mbe;maxi=0;maxj=0;
size=Dimensions[h0][[1]];
For[i=1,i<=size,i++,
For[j=1,j<=size,j++,
If[i!=j,
{\[Alpha]u,h0u,cumu}=PHTrafo[\[Alpha],h0,cum,i,j,a];
If[ CheckPHEnt[\[Alpha]u,h0u],m=MeasureEnt[\[Alpha]u,h0u];
If[m >max,maxi=i;maxj=j;max=m ;maxa=a]
];
{\[Alpha]u,h0u,cumu}=PHTrafo[\[Alpha],h0,cum,i,j,-a];
If[ CheckPHEnt[\[Alpha]u,h0u],m=MeasureEnt[\[Alpha]u,h0u];
If[m >max,maxi=i;maxj=j;max=m ;maxa=-a]
];

];
];
];
If[maxi!=0,
{\[Alpha]u,h0u,cumu}=PHTrafo[\[Alpha],h0,cum,maxi,maxj,maxa] ;
(*Print["i=",mini," j=",minj," d=", MeasureEnt[h0u,h1u]," a=",mina];*)
Return[{\[Alpha]u, h0u,cumu,True}],
(*Print["i=",mini," j=",minj," d=", mbe ," a=",a];*)
Return[{\[Alpha], h0,cum,False}];
];
]


(*
Input: PH distribution;
Returns;
	mininmal correlation;
	D1 for minimal correlation;
	maximal correlation;
	D1 for maximal correlation;

TODO No error handling yet ;
*)
MinMaxLag1Corr[\[Pi]1_,D0_]:=
Module[{D0i,len,\[Lambda],\[Alpha],c1,c2,v1,v2, D1vars, c,vars,ma1,\[Gamma],ma2,ma,d,\[Pi]tr,b,resMinRaw,D1MinCorr,resMaxRaw,D1MaxCorr,minCorr,maxCorr},

	(* Counting c, coeffitients of the expression to minimize*)
	D0i=Inverse[-D0];
	len=Length[D0];
	\[Lambda]=1/\[Pi]1.D0i.ones2[len];
	\[Alpha]=\[Lambda]*\[Pi]1.D0i;
	c1=\[Lambda]^2/(2*\[Lambda]*\[Alpha].D0i.ones2[len]-1);
	c2=1/(2*\[Lambda]*\[Alpha].D0i.ones2[len]-1);
	
	v1=\[Pi]1.MatrixPower[D0i,2];
	v2=D0i.ones2[len];
	
	(* ISSUE to be taken care of!!! *)
	D1vars=Table[Subscript[dToCountLag1Corr, i,j],{i,len},{j,len}];

	(* Column vector of coeffitients *)
	c=c1*Flatten[Transpose[Transpose[{v1}].{v2}]]-c2;
	
	(* Column vector of variables *)
	vars=Flatten[Transpose[D1vars]];

	(* Making m (ma) constraint matrix *)
	ma1=Table[If[Mod[j,len]==Mod[i,len],1,0],{i,len},{j,len^2}];

	\[Gamma]=\[Pi]1.D0i;
	ma2=Table[
		Table[0,{len*(i-1)}]~Join~\[Gamma]~Join~Table[0,{len^2-len*i}],
		{i,len}];

	ma=Join[ma1,ma2];

	(* Making b constraint vector *)
	d=Transpose[{-D0.ones2[len]}];
	\[Pi]tr=Transpose[{\[Pi]1}];
	b=Join[Join[d,\[Pi]tr],Table[{0},{2len}],2];	

	resMinRaw=LinearProgramming[c,ma,b];
	D1MinCorr=Table[resMinRaw[[i+(j-1)*len]],{i,len},{j,len}];
	minCorr=(\[Lambda]*\[Alpha].D0i.D1MinCorr.D0i.ones2[len]-1)/(2*\[Lambda]*\[Alpha].D0i.ones2[len]-1);
	resMaxRaw=LinearProgramming[-c,ma,b];
	D1MaxCorr=Table[resMaxRaw[[i+(j-1)*len]],{i,len},{j,len}];
	maxCorr=(\[Lambda]*\[Alpha].D0i.D1MaxCorr.D0i.ones2[len]-1)/(2*\[Lambda]*\[Alpha].D0i.ones2[len]-1);

	Return[{minCorr,D1MinCorr,maxCorr,D1MaxCorr}];
]



(*
Input: PH distribution;
Returns;
	maximal correlation;
	D1 for maximal correlation;

TODO No error handling yet ;
*)
MaxLag1Corr[\[Pi]1_,D0_]:=
Module[{D0i,len,\[Lambda],\[Alpha],c1,c2,v1,v2, D1vars, c,vars,ma1,\[Gamma],ma2,ma,d,\[Pi]tr,b,resMinRaw,D1MinCorr,resMaxRaw,D1MaxCorr,minCorr,maxCorr},

	(* Counting c, coeffitients of the expression to minimize*)
	D0i=Inverse[-D0];
	len=Length[D0];
	\[Lambda]=1/\[Pi]1.D0i.ones2[len];
	\[Alpha]=\[Lambda]*\[Pi]1.D0i;
	c1=\[Lambda]^2/(2*\[Lambda]*\[Alpha].D0i.ones2[len]-1);
	c2=1/(2*\[Lambda]*\[Alpha].D0i.ones2[len]-1);
	
	v1=\[Pi]1.MatrixPower[D0i,2];
	v2=D0i.ones2[len];
	
	(* ISSUE to be taken care of!!! *)
	D1vars=Table[Subscript[dToCountLag1Corr, i,j],{i,len},{j,len}];

	(* Column vector of coeffitients *)
	c=c1*Flatten[Transpose[Transpose[{v1}].{v2}]]-c2;
	
	(* Column vector of variables *)
	vars=Flatten[Transpose[D1vars]];

	(* Making m (ma) constraint matrix *)
	ma1=Table[If[Mod[j,len]==Mod[i,len],1,0],{i,len},{j,len^2}];

	\[Gamma]=\[Pi]1.D0i;
	ma2=Table[
		Table[0,{len*(i-1)}]~Join~\[Gamma]~Join~Table[0,{len^2-len*i}],
		{i,len}];

	ma=Join[ma1,ma2];

	(* Making b constraint vector *)
	d=Transpose[{-D0.ones2[len]}];
	\[Pi]tr=Transpose[{\[Pi]1}];
	b=Join[Join[d,\[Pi]tr],Table[{0},{2len}],2];	

	resMaxRaw=LinearProgramming[-c,ma,b];
	D1MaxCorr=Table[resMaxRaw[[i+(j-1)*len]],{i,len},{j,len}];
	maxCorr=(\[Lambda]*\[Alpha].D0i.D1MaxCorr.D0i.ones2[len]-1)/(2*\[Lambda]*\[Alpha].D0i.ones2[len]-1);

	Return[{maxCorr,D1MaxCorr}];
]



(*
Input: PH distribution;
Returns;
	mininmal correlation;
	D1 for minimal correlation;
	maximal correlation;
	D1 for maximal correlation;

TODO No error handling yet ;
*)
MaxRangeLag1Corr[\[Pi]1_,D0_]:=
Module[{D0i,len,\[Lambda],\[Alpha],c1,c2,v1,v2, D1vars, D1vars2, c,vars,ma1,\[Gamma],ma2,ma,d,\[Pi]tr,b,resRaw,resMinRaw,D1MinCorr,resMaxRaw,D1MaxCorr,minCorr,maxCorr},

	(* Counting c, coeffitients of the expression to minimize*)
	D0i=Inverse[-D0];
	len=Length[D0];
	\[Lambda]=1/\[Pi]1.D0i.ones2[len];
	\[Alpha]=\[Lambda]*\[Pi]1.D0i;
	c1=\[Lambda]^2/(2*\[Lambda]*\[Alpha].D0i.ones2[len]-1);
	c2=1/(2*\[Lambda]*\[Alpha].D0i.ones2[len]-1);
	
	v1=\[Pi]1.MatrixPower[D0i,2];
	v2=D0i.ones2[len];
	
	(* ISSUE to be taken care of!!! *)
	D1vars=Table[Subscript[dToCountLag1Corr, i,j],{i,len},{j,len}];
	D1vars2=Table[Subscript[dToCountLag1Corr2, i,j],{i,len},{j,len}];

	(* Column vector of coeffitients *)
	c=c1*Flatten[Transpose[Transpose[{v1}].{v2}]]-c2;
	c=Join[c,-c];
	
	(* Column vector of variables *)
	vars=Flatten[Transpose[D1vars]];
	vars=vars~Join~Flatten[Transpose[D1vars2]];

	(* Making m (ma) constraint matrix *)
	ma1=Table[If[Mod[j,len]==Mod[i,len],1,0],{i,len},{j,len^2}];

	\[Gamma]=\[Pi]1.D0i;
	ma2=Table[
		Table[0,{len*(i-1)}]~Join~\[Gamma]~Join~Table[0,{len^2-len*i}],
		{i,len}];

	ma=Join[ma1,ma2];
	ma=Join[
			Join[ma, Table[0,{Length[ma]},{Length[ma[[1]]]}], 2]
			,
			Join[Table[0,{Length[ma]},{Length[ma[[1]]]}], ma, 2]
		];

	(* Making b constraint vector *)
	d=Transpose[{-D0.ones2[len]}];
	\[Pi]tr=Transpose[{\[Pi]1}];
	b=Join[Join[d,\[Pi]tr],Table[{0},{2len}],2];	
	b=Join[b,b];

	resRaw=LinearProgramming[c,ma,b];
	D1MinCorr=Table[resRaw[[i+(j-1)*len]],{i,len},{j,len}];
	minCorr=(\[Lambda]*\[Alpha].D0i.D1MinCorr.D0i.ones2[len]-1)/(2*\[Lambda]*\[Alpha].D0i.ones2[len]-1);
	D1MaxCorr=Table[resRaw[[len*len+i+(j-1)*len]],{i,len},{j,len}];
	maxCorr=(\[Lambda]*\[Alpha].D0i.D1MaxCorr.D0i.ones2[len]-1)/(2*\[Lambda]*\[Alpha].D0i.ones2[len]-1);

	Return[{minCorr,D1MinCorr,maxCorr,D1MaxCorr}];
]


(* Trick of Buchholz *)
(*
Input: PH distribution;
Returns: Different representation of the distribution (every state is exit state);

TODO No error handling yet ;
*)
EntBuchholz[\[Alpha]_,A_,\[Delta]rate_]:=
Module[{Q,Qv,\[CapitalLambda],\[Pi]1,\[Pi]1v,len,i1,i,j,k,l,\[Delta]1,\[Delta],\[Beta],B},
	
	\[Pi]1=\[Alpha];
	Q=A;
	len=Length[Q];

	\[CapitalLambda]=Table[0,{len}];
	For[i1=1,i1<=len,i1++,
		\[CapitalLambda][[i1]]=-A[[i1,i1]];
	];

	\[Pi]1v=\[Pi]1;
	Qv=Q;

	For[i=1,i<=len,i++,
		For[j=i+1,j<=len,j++,
			\[Delta]1=Min[\[Pi]1v[[j]],(\[Pi]1v[[i]]*Qv[[i,j]])/(\[CapitalLambda][[j]]-\[CapitalLambda][[i]]),Min[Table[If[Qv[[k,i]]>0,\[Pi]1v[[i]]*Qv[[k,j]]/Qv[[k,i]],\[Pi]1v[[j]]+1],{k,i-1}]]];
			\[Delta]=\[Delta]rate*\[Delta]1;
			\[Pi]1=\[Pi]1v;
			\[Pi]1[[i]]+=\[Delta];
			\[Pi]1[[j]]-=\[Delta];
			Q=Qv;

			For[k=1,k<=len,k++,
				For[l=1,l<=len,l++,
					If[k==i&&l==j,Q[[k,l]]=Qv[[i,j]]*\[Pi]1v[[i]]/(\[Pi]1v[[i]]+\[Delta])-((\[CapitalLambda][[j]]-\[CapitalLambda][[i]])*\[Delta])/(\[Pi]1v[[i]]+\[Delta])];
					If[k==i&&l!=j&&l!=i,Q[[k,l]]=Qv[[i,l]]*\[Pi]1v[[i]]/(\[Pi]1v[[i]]+\[Delta])+Qv[[j,l]]*\[Delta]/(\[Pi]1v[[i]]+\[Delta])];
					If[k<i&&l==i,Q[[k,l]]=Qv[[k,i]]*(\[Pi]1v[[i]]+\[Delta])/\[Pi]1v[[i]]];
					If[k<j&&l==j&&k!=i,Q[[k,l]]=Qv[[k,j]]-Qv[[k,i]]*\[Delta]/\[Pi]1v[[i]]];
				]
			];
			\[Pi]1v=\[Pi]1;
			Qv=Q;		
		]
	];

	B=Qv;
	\[Beta]=\[Pi]1v;

	Return[{\[Beta],B}];
];


(*
Simulates maxArr arrivals of a MAP, with matrices d0 and d1;
Needs BuTools Utilities;
TODO: check;
*)
MAPSim[d0_,d1_,maxArr_]:=
Module[{size,D0i,P,\[Pi]1,\[Alpha]1,i,j,mod,P01mod,rnd,state,arrnum,rndTime,sumTime,lastTime,arrivals,D0,D1,T\[Pi]},
If[Not[CheckMAP[d0,d1]],Return[{-1}]];
size=Length[d0];
D0=d0;
D1=d1;
D0i=Inverse[-d0];
P=D0i.D1;
\[Pi]1=DTMCSolve[P];
\[Alpha]1=CTMCSolve[D0+D1];



For[i=1,i<=Length[D0],i++,
	mod=(((D0+D1).Table[1,{Length[D0]}])[[i]]-D0[[i,i]])/-D0[[i,i]];
	D1[[i]]=D1[[i]]/mod;
	For[j=1,j<=size,j++,
		D0[[i,j]]=If[i!=j,D0[[i,j]]/mod,D0[[i,i]]];
	]
];

P01mod=Table[0,{size},{2*size}];

(* P01mod *)
For[i=1,i<=size,i++,
	P01mod[[i,1]]=If[i==1,0,D0[[i,1]]/-D0[[i,i]]];
	For[j=2,j<=size,j++,
		P01mod[[i,j]]=If[i==j,P01mod[[i,j-1]],P01mod[[i,j-1]]+D0[[i,j]]/-D0[[i,i]]];
	];
	For[j=size+1,j<=2*size,j++,P01mod[[i,j]]=P01mod[[i,j-1]]+D1[[i,j-size]]/-D0[[i,i]]];
];

(* Initial state *)
For[i=1,rnd>=\[Alpha]1[[i]],i++];
state=i;
arrnum=0;
rndTime=0;
sumTime=0;
lastTime=0;
arrivals=Table[0,{maxArr}];

(* Simulation *)
While[arrnum<maxArr,
	rnd=RandomReal[];
	rndTime=Log[1-rnd]/D0[[state,state]];
	sumTime+=rndTime;
	rnd=RandomReal[];
	For[i=1,rnd>=P01mod[[state,i]],i++];
	state=i;
	If[state>size,
		arrnum++;
		arrivals[[arrnum]]=sumTime-lastTime;
		lastTime=sumTime;
		state-=size;
	];
];
Return[arrivals];

];


(*
Creates Buchholz representation of vector from simple vector representation;
TODO: check;
*)
VectToBuchh[v_]:=
Module[{size,buchh},
size=Length[v];
buchh={{1},{size}}~Join~Transpose[{v}];
Return[buchh];
]


(*
Creates Buchholz representation of matrix from simple matrix representation;
TODO: check;
*)
MatToBuchh[a_]:=
Module[{A,size,toExp,i,j,n},
A=a;
size=Length[A];
toExp={{1},{""},{0},{size},{""}};
For[i=1,i<=size,i++,
	For[j=1;n=0,j<=Length[A[[i]]],j++,If[A[[i,j]]!=0,n++]];
	toExp=toExp~Join~{{n-1,"",A[[i,i]]}};

	For[j=1,j<=size,j++,
	If[A[[i,j]]>0,toExp=toExp~Join~{{j-1,A[[i,j]]}}];
	];
];

Return[toExp];
]


(* Counts first max moments of a table;
TODO: error handling *)
MomsFromSim[arrs_,max_]:=
Module[{size,moms,i},
size=Length[arrs];
moms=Table["",{max}];
For[i=1,i<=max,i++,
moms[[i]]=1/size*Total[arrs^i];
];
Return[moms];
]


(* Counts m11 ... m1j, m21 ... mjj joint moments (j=max);
TODO: error handling *)
JMomsFromSim[arrs_,max_]:=
Module[{size,jmoms,i0,i,j},
size=Length[arrs];
jmoms=Table[0,{max},{max}];
For[i0=1,i0<size,i0++,
	For[i=1,i<=max,i++,
		For[j=1,j<=max,j++,
		jmoms[[i,j]]+=arrs[[i0]]^i*arrs[[i0+1]]^j;
		];
	];
];
jmoms=jmoms/(size-1);

Return[jmoms];
]


(*
Transforms a vector of moments into Buchholz representation;
TODO: error handling
*)
MomsToBuchh[v0_]:=
Module[{v,size,toExp},
size=Length[v0];
v=v0;
toExp={{size}}~Join~Table[{v[[i]]},{i,size}];

Return[toExp];
];


(*
Transforms a matrix of jont moments into Buchholz representation;
TODO: error handling
*)
JMomsToBuchh[m_]:=
Module[{M,size,toExp},
size=Length[m];
M=m;
toExp={{1},{size},{size}}~Join~M;

Return[toExp];
];


(*  *)
BuchhToMat[raw_]:=
Module[{size,D0,line,i,start,end},
size=raw[[4,1]];
D0=Table[0,{size},{size}];

If[Length[raw[[5]]]==2,line=5,line=6];

For[i=1,i<=size,i++,
	For[start=line;end=line+size,line<=end,line++,
		If[line==start,
			D0[[i,i]]=raw[[line,-1]]; end=line+raw[[line,1]],
			D0[[i,raw[[line,1]]+1]]=raw[[line,-1]];
		];
	];
];

Return[D0];
]


BuchhToVect[raw_]:=
Module[{size,v},
size=raw[[2,1]];
v=Table[raw[[2+i,1]],{i,size}];

Return[v];
]


BuchhToJmoms[raw_]:=
Module[{size,jmoms},
size=raw[[3,1]];
jmoms=Table[raw[[3+i,j]],{i,size},{j,size}];

Return[jmoms];
]


PhFitToPH[raw_]:=
Module[{size,\[Alpha],i,j,A},
size=raw[[1,1]];
\[Alpha]=Reverse[raw[[2]]];
A=Table[
	If[i==j,-raw[[3,size-i+1]],If[i==j-1,raw[[3,size-i+1]],0]],
{i,size},{j,size}];

Return[{\[Alpha],A}];
]


(* TODO: as always: checking *)
MinimizeJmomDiff[\[Pi]1_,D0_,jmoms_]:=
Module[{D0i,size,c,\[Lambda],\[Alpha],D1vars,d1var,i,j,jlen,vars,ma1,\[Gamma],ma2,map,d,\[Pi]tr,bp,jmvect,D1opt,cp,ma,resMinRaw,d1optv,b,betavect,const,cost,mincost,D1min},
	(* Computing a few necessary values *)	

	D0i=Inverse[-D0];
	size=Length[D0];
	c=Table[0,{i,size},{j,size}];

	\[Lambda]=1/\[Pi]1.D0i.ones2[size];
	\[Alpha]=\[Lambda]*\[Pi]1.D0i;
	
	(* Do not modify it! *)
	jlen=2;
	
	
	
	(* ISSUE to be taken care of!!! *)
	D1vars=Table[Subscript[d1var, i,j],{i,size},{j,size}];

	(* Column vector of variables *)
	vars=Flatten[Transpose[D1vars]];

	(* Making first part of m constraint matrix (map) *)
	ma1=Table[If[Mod[j,size]==Mod[i,size],1,0],{i,size},{j,size^2}];

	\[Gamma]=\[Pi]1.D0i;
	ma2=Table[
		Table[0,{size*(i-1)}]~Join~\[Gamma]~Join~Table[0,{size^2-size*i}],
		{i,size}];

	map=Join[ma1~Join~ma2,Table[{0},{Length[ma1]+Length[ma2]}],2]~Join~{Table[ If[j!=(Dimensions[ma1][[2]]+1),0,1],{j,Dimensions[ma1][[2]]+1}]};
	
	(* Making first part of b constraint vector (bp) *)
	d=Transpose[{-D0.ones2[size]}];
	\[Pi]tr=Transpose[{\[Pi]1}];
	bp=Join[Join[d,\[Pi]tr],Table[{0},{2size}],2];	

	jmvect=Flatten[ jmoms[[1;;2,1;;2]] ];
	D1opt=Table[0,{2},{2},{2},{2}];
	
	(* Making parts of the cost vector *)
	cp=Table[0.25^(i-1+j-1)*1/jmoms[[i,j]]*i!*j!*Flatten[Transpose[{MatrixPower[D0i,j].ones2[size]}].{\[Pi]1.MatrixPower[D0i,i+1]}],{i,2},{j,2}];

	(* Just to get rid of some nasty brackets *)
	cp=Level[cp,{2}];
	(* Finding the optimal solution for the first 2x2 joint moments *)
	i=Table[0,{4}];
	For[i[[1]]=1,i[[1]]<=2,i[[1]]++,
	For[i[[2]]=1,i[[2]]<=2,i[[2]]++,
	For[i[[3]]=1,i[[3]]<=2,i[[3]]++,
	For[i[[4]]=1,i[[4]]<=2,i[[4]]++,
		ma=map~Join~Table[cp[[n]]~Join~{0},{n,4}];
		(* The last row is for the dummy variable *)
		c=Sum[(-1)^i[[n]]*cp[[n]],{n,4}]~Join~{1};
		betavect=Flatten[ Table[0.25^(i+j-2),{i,2},{j,2}] ];
		(* const is our constant for the dummy variable needed for the goal function *)
		const=Sum[(-1)^(i[[n]]+1)*betavect[[n]],{n,4}];
		(* Second part is for auxiliary variables (equal to 1), third part is for the partitioning of the state space *)
		b=bp~Join~{{const,0}}~Join~Table[{betavect[[n]],(-1)^i[[n]]},{n,4}];

		(* If we don't have a result in the part, fill result with blank *)
		Quiet[resMinRaw=Check[LinearProgramming[c,ma,b],{}]];
		If[Length[resMinRaw]==0,
			D1opt[[ i[[1]],i[[2]],i[[3]],i[[4]] ]]={};
			cost=size^2+1;
			,
			D1opt[[ i[[1]],i[[2]],i[[3]],i[[4]] ]]=Table[resMinRaw[[i+(j-1)*size]],{i,size},{j,size}];
			d1optv=resMinRaw[[1;;size^2]];
			cost=c[[1;;size^2]].d1optv+const;
		];

		If[Sum[i[[n]],{n,4}]==4,
			mincost=Abs[cost];
			D1min=D1opt[[ i[[1]],i[[2]],i[[3]],i[[4]] ]];
			,
			If[Abs[cost]<=mincost,
				mincost=Abs[cost];
				D1min=D1opt[[ i[[1]],i[[2]],i[[3]],i[[4]] ]];
			];
		];

	]
	]
	]
	];
	Return[{mincost,D1min}];
]


End[(* Private *)]

Protect[ones2, RationalToMarkovian2, METoPH2, RndTransf, PH2PHuniv2, MinMaxLag1Corr, MaxLag1Corr, MaxRangeLag1Corr, EntBuchholz, MAPSim, VectToBuchh, MatToBuchh, MomsFromSim, JMomsFromSim, MomsToBuchh,
JMomsToBuchh, BuchhToMat, BuchhToVect, BuchhToJmoms, PhFitToPH, MinimizeJmomDiff];

EndPackage[]
