(* ::Package:: *)

(*
   BuTools_PH Package
*)

(*FONTOS! A RUN PACKAGE CSAK AKKOR FOG JOL MUKODNI, HA ELOTTE*)
(*A PROBA FAJLBAN A DIR, APPENTO, PATH RESZT LEFUTTATOD!!! *)

BeginPackage["BuTools`PH`"];
MomentsFromME::usage = "MomentsFromME [ vector, matrix, k ] -> [ moments ] : Calculates the first k moments of a ME given with a vector-matrix pair"
MomentsFromPH::usage = "MomentsFromPH [ vector, matrix, k ] -> [ moments ] : Calculates the first k moments of a PH given with a vector-matrix pair"
MEFromMoments::usage = "MEFromMoments [ moments ] -> [ vector, matrix ] : Calculates the minimal ME representation based on 2n-1 moments" 
Appie::usage = "Appie [ reducedmoments, print flag ] -> [ matrix ] : Calculates the minimal ME representation with e_1 opening and closing vector based on 2n-1 reducedmoments" 
PH3Canonical::usage = "PH3Canonical [ vector, matrix ] -> [ vector, matrix ] : Calculates the order 3 canonical representation from any order 3 PH representation."
ME3Canonical::usage = "ME3Canonical [ vector, matrix ] -> [ vector, matrix ] : Calculates the order 3 canonical representation from any order 3 vector-matrix representation, if exists"
METoPH::usage = "METoPH [ vector, matrix ] -> [ vector, matrix ] : Converts  matrix exponential vector-matrix pair to phase type vector-matrix pair"
PH2Canonical::usage = "PH2Canonical [ vector, matrix ] -> [ vector, matrix ] : Calculates the order 2 canonical representation from any order 2 vector-matrix representation, if exists" 
APHRepresentation::usage = "APHRepresentation [ vector, matrix ] -> [ vector, matrix ] : Calculates the APH (CF1) representation from any order n vector-matrix representation, if exists " 
MonocyclicRepresentation::usage = "MonocyclicRepresentation [ vector, matrix ] -> [ vector, matrix ] : Calculates the representation of the input ME distribution with Markovian monocyclic generator"
APHFrom3Moments::usage = "APHFrom3Moments [ mom1, mom2, mom3] -> [ vector, matrix ] : Calculates the smallest APH with these 3 moments" 
MEMoments::usage = "MEMomenst [ vector, matrix ] -> [ moments ] : Calculates 2n-1 moments of the ME distribution from the vector-matrix pair of size n" 
MEOrderFromMoments::usage = "MEOrderFromMoments [ moments ] -> [ order ] : Calculates the order of the ME distribution based on its moments using the determinant of the Hankel matrix"
ME3member::usage = "ME3member [ vector, matrix ] -> [ flag ] : Checks if the vector-matrix pari is a member of ME3"
MEContOrder::usage = "MEContOrder [ vector, matrix ] -> [ order ] : Controllability (closing vector) order of the vector-matrix pair"
MEObsOrder::usage = "MEObsOrder [ vector, matrix ] -> [ order ] : Observability (initial vector) order of the vector-matrix pair"
MEObsMinimize::usage = "MEObsminimize [ vector, matrix ] -> [ vector, matrix ] : Minimizies the representation according to the initial vector"
MEContMinimize::usage = "MEContMinimize [ vector, matrix ] -> [ vector, matrix ] : Minimizies the representation according to the closing vector"
MEMinimize::usage = "MEMiminize [ vector, matrix ] -> [ vector, matrix ] : Minimizies the representation according to the initial and the closing vector"
CheckMEPositiveDensity::usage = "CheckMEPositiveDensity [ vector, matrix ] -> [ flag ] : Checks if the vector-matrix pair results in a positive density"
MEDensity::usage = "MEDensity [ vector, matrix, x ]  -> [ density value ] : Gives back the value of the probability density function at point x."


Unprotect[MomentsFromME, MomentsFromPH, MEFromMoments, Appie, 
PH3Canonical, ME3Canonical,
METoPH, PH2Canonical, APHRepresentation, MonocyclicRepresentation, 
APHFrom3Moments, MEMoments, MEOrderFromMoments, ME3member,
MEContOrder, MEObsOrder, MEObsMinimize, MEContMinimize, MEMinimize, MEDensity, CheckMEPositiveDensity];


Begin["`Private`"]

Needs["BuTools`Utilities`"]



MomentsFromME[\[Alpha]_,T_,k_]:=
Module[{i,e,n,res,Ti},
(* If[Not[CheckME[\[Alpha],T]], Return[{0}]]; *)
If[Not[CheckME[\[Alpha],T]], Print["Moments of an invalid ME are computed"] ];
n=Dimensions[T][[1]];
res={};
e=Table[1,{n},{1}];
Ti=Inverse[-T];
For[i=1,i<=k,i=i+1,
res=Append[res,i! (\[Alpha].MatrixPower[Ti,i].e)[[1]]];
];
res
]



MomentsFromPH[\[Alpha]_,T_,k_]:=
Module[{i},
If[Not[CheckPH[\[Alpha],T]], Return[{0}]];
MomentsFromME[\[Alpha],T,k]
]


(* input: vector of reduced moments /\[Mu]_i/i!/ from 1 to 2m-1 *) 
(* output: Appie's ME representation: matrix V and order n *) 
(*calculates the minimal ME representation based on 2n-1 moments *) 
MEFromMoments[mom_]:=
Module[{size,kk,rmom,hh,e1,tt,uu,i,j,tti,uui},
rmom=MomsToReducedmoms[mom];
{kk,size}=Appie[rmom,False] ;

hh=Table[1,{size},{1}];
e1=Table[0,{1},{size}];
e1[[1,1]]=1;

tt=Table[0,{size},{size}];
For[i=1,i<=size,i++,For[j=1,j<=i,j++,tt[[i,j]]= 1;];];
tti=Inverse[tt];
(*Print[tt//MatrixForm,tti//MatrixForm];*)

uu=Table[0,{size},{size}];
For[i=1,i<=size,i++,For[j=i,j<=size,j++,uu[[i,j]]= 1/(size-i+1);];];
uui=Inverse[uu];
(* Print[uu//MatrixForm,uui//MatrixForm]; *)

Return[{(e1)[[1]].tti.uu,-Inverse[uui.tt.kk.tti.uu]}];
] 


Appie[rmom_,print_] :=
(* input: vector of reduced moments /\[Mu]_i/i!/ from 1 to 2m-1 *) 
(* output: Appie's ME representation: matrix V and order n *) 
Module[ {rm,m, i,j,f,y,dd,nold,yold,k,q,d,\[Alpha],\[Beta],n,\[Rho],kk,ind,inc},
     m=Dimensions[rmom][[1]];
If[Mod[m,2]==0,
Print["Even number of moments, the last one is dropped !!"];
rm=Drop[rmom,-1];m=m/2,rm=rmom;m=Ceiling[m/2]
];
rm=Prepend[rm,1];
f=Table[0,{2m},{1}];
f[[1,1]]=1;
y=Table[0,{2m},{1}];
dd=Table[0,{2m},{2m}];
n=0;
k=0;
q=1;
d=Table[0,{m}];
\[Alpha]=Table[0,{m},{m}];
\[Beta]=Table[0,{m}];
For[i=2,i<=2m,dd[[i,i-1]]=1;i++];
For[i=1,i<=2m,
(*THEN*)
\[Rho]=FullSimplify[q (rm.f)[[1]] ];
nold=n;
n=nold+1;
yold=y;
If[print,Print["iter ",i," \[Rho]=",\[Rho]," n=",n, " yold=",yold]];
If[n>0 &&\[Rho]!=0 ,
(*THEN*)
If[k>0,\[Beta][[k]]=\[Rho]/(rm[[1+1]])^(d[[k]]+n-1)];
k=k+1;
d[[k]]=n;
n=-n;
q=q/\[Rho];
y=dd.f;
If[print,Print[i," then: \[Beta]=",\[Beta]," k=",k," n=",n," q=",q ]],
(*ELSE*)
If[print,Print[i," else: n=",n]];
If[n<=0,
j=nold+d[[k]]+1;
\[Alpha][[k,j]]=\[Rho]/(rm[[1+1]])^(j-1);
If[print,Print[i ," else if: k+1=",k+1," j+1=",j+1, " \[Alpha]=",\[Alpha]]]
],
(*UNDEFINED /in case of symbolic analysis/ *)
If[n>0 ,
(*THEN*)
If[k>0,\[Beta][[k]]=\[Rho]/(rm[[1+1]])^(d[[k]]+n-1)];
k=k+1;
d[[k]]=n;
n=-n;
q=q/\[Rho];
y=dd.f;
If[print,Print[i," undefined then: \[Beta]=",\[Beta]," k=",k," n=",n," q=",q ]],
(*ELSE*)
If[print,Print[i," else: n=",n]];
If[n<=0,
j=nold+d[[k]]+1;
\[Alpha][[k,j]]=\[Rho]/(rm[[1+1]])^(j-1);
If[print,Print[i ," undefined else if: k+1=",k+1," j+1=",j+1, " \[Alpha]=",\[Alpha]]]
]
]
];
f= dd.f-\[Rho] yold;
If[print,Print["f=",f]];
i++];

If[print,Print["m=",m," \[Beta]=",\[Beta]," \[Alpha]=",\[Alpha]," d=",d]];
If[Sum[d[[i]],{i,1,m}]!=m,Print["Insufficient matrix order !!!!"]];

kk=Table[0,{m},{m}];
kk[[1,1]]=rm[[2]];
For[i=1,i<=m-1,kk[[i,i+1]]=rm[[2]];i++];

ind=d[[1]];
For[i=2,i<=m,
If[ind<m,
inc=d[[i]];
If[(ind=ind+inc)<=m,
kk[[ind,ind-inc-d[[i-1]]+1]]=\[Beta][[i-1]];
For[j=1,j<=inc,
kk[[ind,ind-j+1]]=\[Alpha][[i,j]];
j++];
];
];
i++];
{kk,m}
	]


PH3Canonical[\[Alpha]_,T_]:=
Module[{i},
If[Not[CheckPH[\[Alpha],T]], Return[{0}]];
ME3Canonical[\[Alpha],T]
]


ME3Canonical[v_,h0_] :=
Module[ {ii,\[Lambda],a0,a1,a2,a3,x1,x2,x3,x13,\[Gamma]u,\[Gamma]0,\[Gamma]l,\[Phi],f,p1,p2,p3,p,lower1,e},

If[Not[CheckME[v, h0]], Return[{0}]];
If[Dimensions[h0][[1]] !=3,Print["Dimension \[NotEqual] 3 !!! "];Return[] ];

e={{1},{1},{1}};
ii=IdentityMatrix[3];
\[Lambda]=-Eigenvalues[h0];
Print["eigenvalues: ",\[Lambda]];

If[(Re[\[Lambda][[3]]]<0 )||Abs[Im[\[Lambda][[3]]]]>0, 
 Print["eigenvalue error!!!"]; Return[{v,h0}]
];

a0=\[Lambda][[1]] \[Lambda][[2]] \[Lambda][[3]] ;
a1=\[Lambda][[1]] \[Lambda][[2]] + \[Lambda][[1]] \[Lambda][[3]] + \[Lambda][[2]] \[Lambda][[3]];
a2=\[Lambda][[1]] + \[Lambda][[2]] +  \[Lambda][[3]];
{a0,a1,a2}=Re[{a0,a1,a2}];

f[x_]:= x^3+a2  x^2+a1 x+a0;
\[Gamma]u=(a2+2 Sqrt[a2^2-3 a1])/3;
\[Gamma]0=(a2+ Sqrt[a2^2-3 a1])/3;
If[\[Lambda][[1]]\[Element] Reals,\[Gamma]l=\[Lambda][[1]],\[Gamma]l=\[Gamma]0];
lower1=- (v.h0.h0.e)[[1]]/(v.h0.e)[[1]];
Print["\[Gamma]u=",\[Gamma]u,", \[Gamma]l=",\[Gamma]l,", lower1=",lower1];
If[lower1>\[Gamma]l ,\[Gamma]l=lower1];

If[\[Gamma]l>\[Gamma]u ,Print["Max(\[Gamma]l,lower) > \[Gamma]u !!!!!!"];x1=\[Gamma]u,x1=\[Gamma]l];
Print["Degree of freedom: \[Gamma]u - Max(\[Gamma]l,lower)=",\[Gamma]u - Max[\[Gamma]l,lower1]]; 

x13=-f[-x1]/(a0-f[-x1]) x1;
x2=(a2-x1+Sqrt[(a2-x1)^2-4(x1^2 - a2 x1 +a1)])/2;
x3=(a2-x1-Sqrt[(a2-x1)^2-4(x1^2 - a2 x1 +a1)])/2;
Print["x1=",x1,", x13=",x13,", x2=",x2,", x3=",x3];
{x13,x2,x3}=Re[{x13,x2,x3}];
(* Print["x1=",x1,", x13=",x13,", x2=",x2,", x3=",x3];*)
If[x13>x1,Print["x13 > x1 !!!!"] ];
If[x13<0,Print["x13 < 0 !!!!"] ];
p1=- h0.{{1},{1},{1}}/(x1-x13);p1=Transpose[p1][[1]];
p2=-(x1 ii + h0).h0.{{1},{1},{1}}/(x1-x13)/x2;p2=Transpose[p2][[1]];
p3=-(x2 ii + h0).(x1 ii + h0).h0.{{1},{1},{1}}/(x1-x13)/x2/x3;p3=Transpose[p3][[1]];
p=Transpose[{p1,p2,p3}];
Return[{v.p,{{-x1,0,x13},{x2,-x2,0},{0,x3,-x3}}}];
];




METoPH[\[Alpha]_,A_]:= 
Module[{size},
size=Dimensions[A][[1]];
X={\[Alpha],A};
Return[RationalToMarkovian[ X,size, PHTransFunc, PHEvalFunc, PHCheckFunc]];
];


PHCheckFunc[X_] :=CheckPH[X[[1]],X[[2]]];

PHTransFunc[X_,i_,j_,a_] :=
Module[ {bb,size},
size=Dimensions[X[[2]]][[1]];
bb=IdentityMatrix[size];
bb[[i,j]]=a;
bb[[i,i]]=1-a;
Return[{X[[1]].bb,Inverse[bb].X[[2]].bb}]
];
PHEvalFunc[X_,Error_] :=
Module[ {i,j,sum,size,\[Alpha],A,a},
size=Dimensions[X[[1]]][[1]];
\[Alpha]=X[[1]];
A=X[[2]];
a=-A.Table[1,{size}]; (* exit rate *)
sum=0;
For[i=1,i<=size,i++,
sum+=Error[\[Alpha][[i]]  ];];
For[i=1,i<=size,i++,
For[j=1,j<=size,j++,
If[i!=j,sum+=Error[A[[i,j]]  ] ];];];
For[i=1,i<=size,i++,
sum+=Error[a[[i]]  ];];
Return[sum];
];



PH2Canonical[\[Alpha]_,A_] :=
Module[ {\[Alpha]v,Av,W,w,k1,k2,Ones, size1,size2},

 If[Not[CheckME[\[Alpha],A]], Return[{{0},{{0}}}]]; 

If[Dimensions[A][[1]] !=2,Print["Dimension \[NotEqual] 2 !!! "];Return[{{0},{{0}}}] ];

{\[Alpha]v,Av}=APHRepresentation[\[Alpha],A] ;

If[Not[CheckPH[\[Alpha]v,Av]],Print["Not a PH(2) !!! "];Return[{{0},{{0}}}]]; 

Return[{\[Alpha]v,Av}];
];




(*  Calculates the APH (CF1) representation 
from any order n vector-matrix representation, if exists *)
APHRepresentation[\[Alpha]_,A_] :=
Module[ {\[Lambda],mx,W,w,k1,k2,Ones, size1,size2},

 If[Not[CheckME[\[Alpha],A]], Return[{{0},{{0}}}]]; 

\[Lambda]=Eigenvalues[A//N];
(* Print["eigenvalues: ",\[Lambda]]; *)

If[Max[Abs[Im[ \[Lambda]]]]>0, 
Print["Complex eigenvalue!!!"]; Return[{{0},{{0}}}]
];

\[Lambda]=Sort[\[Lambda],Greater];
mx=DiagonalMatrix[\[Lambda]];
\[Lambda]=Drop[\[Lambda],-1];
mx+=DiagonalMatrix[-\[Lambda],1];

size1=size2=Dimensions[A][[1]] ;
W=Table[Table[Subscript[w, k2,k1],{k1,1,size1}],{k2,1,size2}];
Ones=Table[{1},{size2}];
W=W/.Solve[{W.mx==A.W,W.Ones==Ones}, Flatten[W]][[1]];

Return[{\[Alpha].W,mx}];
];



MonocyclicRepresentation[\[Alpha]_,A_,exitProb_:0]:=
Module[{fel,\[Beta],B},
{fel,\[Beta],B}=MocanuMonocyclic[\[Alpha],A,exitProb];
Return[{\[Beta],B}];
];



MocanuMonocyclic[\[Alpha]_,A_,exitProb_:0]:=Module[{evl,fel,fel2,\[Beta],B,ret},
evl=Eigenvalues[A];
fel=BuildErlangs[evl];

fel2=BuildFEs[evl];
(*Print["fel2=",fel2[[1]]];*)

fel=Join[fel,fel2];
(*Print["fel=",fel];*)

fel=Sort[fel,SortFunction];
B=BuildMatrix[fel,exitProb];
(*FindInitialVector is replaced with RepTrafo *)
(* \[Beta]=FindInitialVector[\[Alpha],A,B]; *)
\[Beta]=RepTrafo[\[Alpha],A,B]; 
ret={fel,\[Beta],B};

(*
If[Min[\[Beta]]<0,Print["Initial vector contains negative entries; starting iterative growth to get rid of them."];
ret=EliminateNegativeEntries[\[Alpha],A,\[Beta],fel];
];
*)
Return[ret];
];



BuildErlangs[evl_]:=Module[{realPoles,currErl,k,\[Lambda],ret},
realPoles=Select[evl,Im[#]==0&];
ret={};
While[Length[realPoles]>0,
\[Lambda]=realPoles[[1]];
currErl=Select[realPoles,#==\[Lambda]&];
ret=AppendTo[ret,{Length[currErl],-\[Lambda],0,\[Lambda]}];
realPoles=Select[realPoles,#!=\[Lambda]&];
];
Return[ret];
];





BuildFEs[evl_]:=Module[{complexPoles,ret,\[Lambda]0,\[Lambda]c,\[Lambda],n,z,currPair},
complexPoles=Select[evl,Im[#]!=0&];
\[Lambda]0=Max[Select[evl,Im[#]==0&]];
ret={};
While[Length[complexPoles]>0,
\[Lambda]c=complexPoles[[1]];
(*
currPair=Select[complexPoles,Re[#]==Re[\[Lambda]c]&];
If[Length[currPair]!=2,
Print["Something funny happend. Please investigate BuildFEs."];
Return;
];
(*Print[\[Lambda]c,\[Lambda]0,currPair];*)
*)
n=FindRelativeOrder[\[Lambda]c,\[Lambda]0];
If[n>20,Print["Large Erlang feedback block !!! Order=",n," !!!!"]];

\[Lambda]=\[Lambda]ForFE[\[Lambda]c,n];
z=zForFE[\[Lambda]c,n];
ret=AppendTo[ret,{n,\[Lambda],z,\[Lambda]c}];(* We store \[Lambda]c here, too, for later re-computation of the FE block. *)
(*complexPoles=Select[complexPoles,Re[#]!=Re[\[Lambda]c]&];*)
complexPoles=Drop[complexPoles,2];
];
Return[ret];
];




SortFunction[fe1_,fe2_]:=Module[{ev1,ev2},
ev1=FeEigenvalues[fe1];ev2=FeEigenvalues[fe2];
(*Print["ev1=",ev1," ev2=",ev2];*)
If[ev1[[1]]>ev2[[1]],Return[True]];
If[ev1[[1]]<ev2[[1]],Return[False]]; 
If[fe1[[2]]<fe2[[2]],Return[True]];
If[fe1[[2]]>fe2[[2]],Return[False]];
Return[fe1[[1]]<=fe2[[1]]];
];




FindRelativeOrder[\[Lambda]c_,\[Lambda]r_]:=Module[{m,a,b,\[Lambda]},
a=Abs[Re[\[Lambda]c]];b=Abs[Im[\[Lambda]c]];\[Lambda]=Abs[Re[\[Lambda]r]];
Return[Ceiling[(2 \[Pi])/(\[Pi]- 2 ArcTan[b/(a-\[Lambda])])]] 
(* old version 
m=3;
While[a-\[Lambda]r<=b Tan[\[Pi]/m],
m++;
];
Return[m]; *)
];

(* Compute the rate \[Lambda] for an FE block of length n which handles the complex pole \[Lambda]c. *)
\[Lambda]ForFE[\[Lambda]c_,n_]:=Module[{a,b},
a=Abs[Re[\[Lambda]c]];b=Abs[Im[\[Lambda]c]];
Return[1/2 (2a-b Tan[\[Pi]/n]+b Cot[\[Pi]/n])];
];

(* Compute the feedback probability z for an FE block of length n that handles the complex pole \[Lambda]c. *)
zForFE[\[Lambda]c_,n_]:=Module[{a,b,z},
a=Abs[Re[\[Lambda]c]];b=Abs[Im[\[Lambda]c]];
z=(1-(2(a-b Tan[\[Pi]/n]))/(2a-b Tan[\[Pi]/n]+b Cot[\[Pi]/n]))^n;
Return[z];
];

(* For FE block Eigenvalue computation *)
\[Theta][n_,k_]:=2\[Pi] k/n;
s[k_,b_,\[Lambda]_,z_]:=\[Lambda] (z^(1/b) Cos[\[Theta][b,k-1]]-1)+z^(1/b) \[Lambda] Sin[\[Theta][b,k-1]]I;
(* FE data structure: (b,\[Lambda],z) *)

(* Compute the Eigenvalues of an FE block. *)
FeEigenvalues[fe_]:=Module[{k},
Return[Table[s[k,fe[[1]],fe[[2]],fe[[3]]],{k,1,fe[[1]]}]];
];

(* Sort function for a list of FE blocks. I think this is what Mocanu meant.
Note the format of FE blocks: {b, \[Lambda], z, \[Lambda]c}, where b is the length, \[Lambda] is the rate, z is the feedback probability, and \[Lambda]c is the complex pole for which this FE block was generated. *)
SortFunction[fe1_,fe2_]:=Module[{ev1,ev2},
ev1=FeEigenvalues[fe1];ev2=FeEigenvalues[fe2];
(*Print["ev1=",ev1," ev2=",ev2];*)
If[ev1[[1]]>ev2[[1]],Return[True]];
If[ev1[[1]]<ev2[[1]],Return[False]];
If[fe1[[2]]<fe2[[2]],Return[True]];
If[fe1[[2]]>fe2[[2]],Return[False]];
Return[fe1[[1]]<=fe2[[1]]];
];




BuildMatrix[feList_,exitProb_:0]:=Module[{ret,k,\[CapitalLambda],n,x,y,z,\[Lambda]},
(* Build the diagonals. *)
\[CapitalLambda]=Flatten[Table[Table[feList[[k,2]],{feList[[k,1]]}],{k,Length[feList]}]];
n=Length[\[CapitalLambda]];
ret=SparseArray[{Band[{1,1}]->-\[CapitalLambda],Band[{1,2}]->Take[\[CapitalLambda],n-1]}];
(* Insert the feedback loop entries. *)
k=1;x=1;
While[k<=Length[feList],
y=x+feList[[k,1]]-1;
If[feList[[k,3]]>0,
\[Lambda]=feList[[k,2]];
z=feList[[k,3]];
ret[[y,x]]=\[Lambda] z;
If[y<n,ret[[y,y+1]]=(1-z) \[Lambda]];
];
If[y<n,ret[[y,y+1]]= ret[[y,y+1]] (1-exitProb)];
x+=feList[[k,1]];
k++;
];
Return[ret];
];


(* Compute the initial vector \[Beta] by solving Subscript[F, \[Alpha],A](s)==Subscript[F, \[Beta],B](s) for s=1,...,n and \[Beta].1==1. This vector might be negative. *)
FindInitialVector[\[Alpha]_,A_,B_]:=Module[{SE,n,\[Beta],k,s,ret,Sol},
n=Length[B];
(*\[Beta]=Table[,{k,n}];*)
Block[{},
\[Beta]=Table[Symbol["\[Beta]"<>ToString[k]],{k,n}]; (* This is a nasty hack. *)
SE=Table[LST[\[Alpha],A,s]==LST[Array[\[Beta],n],B,s],{s,1,n}];
SE=AppendTo[SE,Sum[\[Beta][[k]],{k,1,n}]==1];
(*Print[SE];*)
Sol=Solve[SE,Array[\[Beta],n]];
ret=Table[Sol[[1,k,2]],{k,n}]; (* This is a nasty hack. *)
];
Return[ret];
];

(*FindInitialVector is replaced with RepTrafo !!!!! *)
RepTrafo[\[Alpha]_,A_, G_]:=
(* inut:
initial vector \[Alpha],
matrix A of an ME distribution
generator matrix G of the same ME,
output: 
the initial vector associated with G
  *)
Module[{W, size1,size2,\[Gamma],w,Ones1,Ones2},

size1=Dimensions[A][[1]];
size2=Dimensions[G][[1]];

Ones1=Table[{1},{size1}];
Ones2=Table[{1},{size2}];

W=Table[Subscript[w, k1,k2],{k1,1,size1},{k2,1,size2}];
W=W/.NSolve[{W.G==A.W,W.Ones2==Ones1}, Flatten[W]][[1]];

Return[\[Alpha].W];

];


LST[\[Alpha]_,Q_,s_]:=(1-Sum[\[Alpha][[i]],{i,1,Length[\[Alpha]]}])+\[Alpha].Inverse[(s IdentityMatrix[Length[\[Alpha]]]-Q)].(-Q.Table[{1},{Length[\[Alpha]]}]);



Density[\[Alpha]_,Q_,x_]:=\[Alpha].MatrixExp[Q x].(-Q.Table[{1},{Length[\[Alpha]]}]);


APHFrom3Moments[m1_,m2_,m3_]:=Module[{mym1,n2,n3,\[Alpha]m,A},
If[Det[({
 {1, m1},
 {m1, m2}
})]<=0,Print["Non-positive variance"]; Return[{{0},{{0}}}]];
If[Det[({
 {m1, m2},
 {m2, m3}
})]<=0,Print["No positive distribution with such moments"]; Return[{{0},{{0}}}]];
{mym1,n2,n3}=MomsToNormMoms[{m1,m2,m3}];
{{\[Alpha]m},A}=MOM3Construct[mym1,n2,n3];
Return[{\[Alpha]m,A}];
]



(* Construct minimal PH dist with normalized moments n1,n2,n3  *)
MOM3Construct[n1_,n2_,n3_]:=Module[{n,b1,b2,a1,a2,c0,c1,c2,c3,c4,K1,K2,K3,K4,K5,K6,P1,P2,P3,P41,P42,m1,m2,m3,p1,p2,lambda1,lambda2,mu1,mu2,result,fs,K7,K8,K9,i1,ff,a,p,lambda,mu},
n=2;
While[(n+1)/n>n2 || MOM3lowerbound[n,n2]>n3||MOM3upperbound[n,n2]<n3,n=n+1];
(*Print[(n2+n n2-n n3)];
Print[12 n2^2(n+1)+16n3(n+1)+n2(n(n3-15)(n3+1)-8(n3+3))];*)
If[n3<=2n2-1,
Print["Below the Erlang-line: ErlangExp"];
b1=(4+n-n n3)/(2(n2+n n2-n n3))+(Sqrt[n] \[Sqrt](12 n2^2(n+1)+16n3 (n+1)+n2 (n (n3-15)(n3+1)-8(n3+3))))/(2Sqrt[n2](n2+n n2-n n3));
b2=(4+n-n n3)/(2(n2+n n2-n n3))-(Sqrt[n] \[Sqrt](12 n2^2(n+1)+16n3 (n+1)+n2 (n (n3-15)(n3+1)-8(n3+3))))/(2Sqrt[n2](n2+n n2-n n3));
a1=(( b1 n2-2)(n-1)b1)/((b1-1)n);
a2=(( b2 n2-2)(n-1)b2)/((b2-1)n);
p1=(b1-1)/a1;
p2=(b2-1)/a2;
lambda1=(p1 a1+1)/n1;
lambda2=(p2 a2+1)/n1;
(* Print["b1: ",b1," a1: ",a1," p1: ",p1," lambda1: ",lambda1];
Print["b2: ",b2," a2: ",a2," p2: ",p2," lambda2: ",lambda2]; *)
mu1=p1/(n1-1/lambda1) (n-1);
mu2=p2/(n1-1/lambda2) (n-1);
If[0<= p1<= 1&&lambda1>0&&mu1>0,
(*  Print["+ is good for ",n2,",",n3]; *)
Return[MOM3ErlangExp[n,p1,lambda1,mu1]];];
If[0<= p2<= 1&&lambda2>0&&mu2>0,
(* Print["- is good ",n2,",",n3]; *)
Return[MOM3ErlangExp[n,p2,lambda2,mu2]];];
result={n,p1,lambda1,mu1,p2, lambda2,mu2};
,
Print["Above the Erlang-line:    ExpErlang"];
c4=n2 (3n2-2n3)(n-1)^2;
c3=2n2 (n3-3)(n-1)^2;
c2=6(n-1)(n-n2);
c1=4n (2-n);
c0=n (n-2);
fs=NSolve[c4 f^4+c3 f^3+c2 f^2+c1 f+c0==0,f];

K1=n-1;
K2=n-2;
K3=3n2-2n3;
K4=n3-3;
K5=n-n2;
K7=(1+n2-n3);
K8=n+n2-n n2;
K9=3+3 n2^2+n3-3 n2 n3;
K6=108 K1^2 (4 K2^2 K3 n^2 n2+K1^2 K2 K4^2 n n2^2+4 K1 K5 (K5^2-3 K2 n n2 K7)+\[Sqrt](-16 K1^2 K8^6+(4 K1 K5^3+K1^2 K2 K4^2 n n2^2+4 K2 n n2 (K4 n^2-3 n2 K7+n K9))^2));
P1=-K4/(2 K3);
P2= \[Sqrt](K4^2/(4K3^2)-K5/(K1 K3 n2)+(2^(1/3) (3 K5^2+K2 (K3+2 K4) n n2))/(K3 K6^(1/3) n2)+K6^(1/3)/(3 2^(7/3) K1^2 K3 n2));
P3=(6 K1 K3 K4 K5+4 K2 K3^2 n- K1^2 K4^3 n2)/(4 K1^2 K3^3 n2 P2);
P41=  \[Sqrt]( K4^2/(2 K3^2)-(2 K5)/(K1 K3 n2)- 2^(1/3) (3 K5^2+K2 (K3+2 K4) n n2)/(K3 K6^(1/3) n2)-K6^(1/3)/(3 2^(7/3) K1^2 K3 n2)-P3);
P42= \[Sqrt](K4^2/(2 K3^2)-(2 K5)/(K1 K3 n2)-(2^(1/3) (3 K5^2+K2 (K3+2 K4) n n2))/(K3 K6^(1/3) n2)-K6^(1/3)/(3 2^(7/3) K1^2 K3 n2)+P3);
(* Print[P1-P2-P41];
Print[P1-P2+P41];
Print[P1+P2-P42];
Print[P1+P2+P42]; *)
fs={P1-P2-P41,P1-P2+P41,P1+P2-P42,P1+P2+P42};
(* Print[fs]; *)
result={n};
For[i1=1,i1<=4,i1=i1+1,
ff=fs[[i1]];
a=(2(ff-1)(n-1))/((n-1)(n2 ff^2-2ff+2)-n);
p=N[(ff-1)a];
lambda=N[(a+p)/n1];
mu=N[(n-1)/(n1-p/lambda)];
result=Append[result,p];
result=Append[result,lambda];
result=Append[result,mu];
p=BuToolsReal[p,12];
lambda=BuToolsReal[lambda,12];
mu=BuToolsReal[mu,12];
If[Element[p,Reals]&&Element[lambda,Reals]&&Element[mu,Reals],
If[0<= p<= 1&&lambda>0&&mu>0,(* Print[i1, ". is good for ",n2,",",n3]; *)
Return[MOM3ExpErlang[n,p,lambda,mu]];
];
];
];
];
Print["Unexpected termination of MOM3Construct!"];
result
];


MOM3ExpErlang[n_,p1_,lambda1_,mu1_]:= Module[{AA,TT,resTT,FF,m1,m2,m3},
TT=-mu1 IdentityMatrix[n];
For[i=2,i<n,i=i+1,TT[[i,i+1]]=mu1];
TT[[1,1]]=-lambda1;
TT[[1,2]]=lambda1;
AA=Table[0,{i,1,1},{j,1,n}];
AA[[1,1]]=p1;
AA[[1,2]]=1-p1;
(* Print[MatrixForm[AA]]; *)
(* Print[MatrixForm[TT]]; *)
resTT=TT;
TT=Inverse[-TT];
FF=Table[1,{i,1,n},{j,1,1}];
m1=AA.(TT).FF;
m2=2 AA.(TT).(TT).FF;
m3=6 AA.(TT).(TT).(TT).FF;
(* Print[{m1,m2/m1/m1,m3/m1/m2}]; *)
{AA,resTT}
];



MOM3ErlangExp[n_,p1_,lambda1_,mu1_]:= Module[{AA,TT,resTT,FF,m1,m2,m3},
TT=-mu1 IdentityMatrix[n];
For[i=1,i<n,i=i+1,TT[[i,i+1]]=mu1];
TT[[n,n]]=-lambda1;
AA=Table[0,{i,1,1},{j,1,n}];
AA[[1,1]]=p1;
AA[[1,n]]=1-p1;
(* Print[MatrixForm[AA]]; *)
(* Print[MatrixForm[TT]]; *)
resTT=TT;
TT=Inverse[-TT];
FF=Table[1,{i,1,n},{j,1,1}];
m1=AA.(TT).FF;
m2=2 AA.(TT).(TT).FF;
m3=6 AA.(TT).(TT).(TT).FF;
(* Print[{m1,m2/m1/m1,m3/m1/m2}]; *)
{AA,resTT}
]



MOM3upperbound[n_,n2_]:=Module[{ub,u},
u=1/(n^2 n2) (2(n-2)(n n2-n-1)Sqrt[1+(n (n2-2))/(n-1)]+(n+2)(3n n2-2n-2));
ub=Which[
n2<(n+1)/n,-Infinity,
n2<= n/(n-1),u,
True,Infinity]];



(* lower and upper bound of 3rd normalized moments for n-phase PH distribution with 2nd normalized moment n2 *)
MOM3lowerbound[n_,n2_]:=Module[{lb,l,p,a,x},
l=((3+a)(n-1)+2a)/((n-1)(1+a p))-(2a (n+1))/(2(n-1)+a p (n a+2n-2));
a=(n2-2)/(p (1-n2)+Sqrt[p^2+(p n (n2-2))/(n-1)]);
p=((n+1)(n2-2))/(3n2 (n-1)) ((-2Sqrt[n+1])/Sqrt[-3n n2+4n+4]-1);
x=n2;
lb=Which[
n2<(n+1)/n,Infinity,
n2< (n+4)/(n+1),l,
True,(n+1)/n n2]]


MEMoments[\[Alpha]_,A_]:=
Module[{size,moms,i},
size= Dimensions[A][[1]];
moms=Table[i!  \[Alpha].MatrixPower[Inverse[-A],i].Table[1,{size}],{i,1,2 size -1}];
Return[moms];
];



MEOrderFromMoments[moms_]:=
Module[{momsize,size,rmoms,i,j,n,m,hankel,\[Epsilon]}, 
\[Epsilon]=10^-10;
momsize= Dimensions[moms][[1]];
size= Floor[(momsize+1)/2];
rmoms=MomsToReducedmoms[moms];
rmoms=Prepend[rmoms,1];
For[n=1,n<=size,n++,
hankel=Table[rmoms[[i+j-1]],{i,1,n},{j,1,n}];
Print["n=",n,", det hankel=",Det[hankel], ", precision=",\[Epsilon]//N];
If[ (* Det[hankel]==0 *) Abs[Det[hankel]] < \[Epsilon] && n==size, Return[n-1]];

If[(* Det[hankel]==0 *) Abs[Det[hankel]] <\[Epsilon],
For[m=1,m<= momsize- 2n+1,m++,
hankel=Table[rmoms[[i+j-1+m]],{i,1,n},{j,1,n}];
If[(* Det[hankel]!=0 *) Abs[Det[hankel]] > \[Epsilon],Print["Hankel matrix of r(0)-r(",2n-1,") is 0, but the Hankel matrix of r(",m,")-r(",2n-1+m,") is not 0 !!"]; Return[n];
];
]; (* end For *)
Return[n-1];
];
];
Return[size];
];


ME3member[\[Alpha]_,A_]:=Module[{res},
If[Dimensions[\[Alpha]]!={3}||Dimensions[A]!={3,3},
If[BuToolsVerbose,Print["ME3member: Dimension is not 3"]];Return[False]
];
If[!CheckME[\[Alpha],A],Return[False]];
res=Mx2ME3Rep[\[Alpha],A,BuToolsVerbose] ;
Return[res[[-1]]];
];


Mx2ME3Rep[v_,h0_,print_]:=Module[{res,h,\[Lambda],aa,bb,a0,a1,a2,a3,ar,ac,\[Lambda]1,\[Lambda]2,\[Lambda]3,\[Lambda]r,\[Lambda]c,\[Omega],\[Phi],ft,fvt,t,tmin,tmax,rrr},
If[Dimensions[h0][[1]]!=3,Print["Dimension \[NotEqual] 3 !!! "];Return[res={0,False}]];
h={{1},{1},{1}};If[(v.h)[[1]]<0||(v.h)[[1]]>1,Print["Invalid mass at 0 !!! "];Return[res={0,False}]];
\[Lambda]=-Eigenvalues[h0];If[print,Print["eigenvalues: ",\[Lambda]]];
If[Re[\[Lambda][[3]]]<0||Abs[Im[\[Lambda][[3]]]]>0,If[print,Print["eigenvalue error!!!"]];Return[res={0,False}];];
{aa,bb}=JordanDecomposition[h0];If[\[Lambda][[1]]\[Element]Reals&&\[Lambda][[1]]!=\[Lambda][[2]]&&\[Lambda][[2]]!=\[Lambda][[3]],If[print,Print["CASE1: 3 real eigenvalues"]];
{a1,a2,a3}=-v.aa Transpose[Inverse[aa].h0.h][[1]];{\[Lambda]1,\[Lambda]2,\[Lambda]3}=\[Lambda];{a1,a2,a3}={a3,a2,a1};{\[Lambda]1,\[Lambda]2,\[Lambda]3}={\[Lambda]3,\[Lambda]2,\[Lambda]1};
If[print,Plot[a1 E^(-t \[Lambda]1)+a2 E^(-t \[Lambda]2)+a3 E^(- t \[Lambda]3),{t,0,5}];
Print["at t=1:",a1 E^(-t \[Lambda]1)+a2 E^(-t \[Lambda]2)+a3 E^(-t \[Lambda]3)/.t->1];
Print["{a1=",a1,", \[Lambda]1=",\[Lambda]1,", a2=",a2,", \[Lambda]2=",\[Lambda]2,", a3=",a3,", \[Lambda]3=",\[Lambda]3,"}"];
];
If[a1+a2+a3<0,Print["negative density at 0!!!"];Return[res={1,a1,\[Lambda]1,a2,\[Lambda]2,a3,\[Lambda]3,False}];];
If[a1<0,Print["negative density at \[Infinity]!!!"];Return[res={1,a1,\[Lambda]1,a2,\[Lambda]2,a3,\[Lambda]3,False}];];
If[a2<-(a1 (\[Lambda]3-\[Lambda]1))/(\[Lambda]3-\[Lambda]2)&&a3<(a1 (\[Lambda]2-\[Lambda]1) (-(a2 (\[Lambda]3-\[Lambda]2))/(a1 (\[Lambda]3-\[Lambda]1)))^((\[Lambda]3-\[Lambda]1)/(\[Lambda]2-\[Lambda]1)))/(\[Lambda]3-\[Lambda]2),Print["negative density on (0,\[Infinity])!!!"];
Return[res={1,a1,\[Lambda]1,a2,\[Lambda]2,a3,\[Lambda]3,False}];];
Return[res={1,a1,\[Lambda]1,a2,\[Lambda]2,a3,\[Lambda]3,True}];];
If[\[Lambda][[1]]\[Element]Reals&&\[Lambda][[1]]==\[Lambda][[2]]&&\[Lambda][[2]]==\[Lambda][[3]],If[print,Print["CASE2: 1 real eigenvalue"]];
\[Lambda]1=\[Lambda][[1]];a0=-(v.h0.h)[[1]];a1=-((v.aa).{{0,1,0},{0,0,1},{0,0,0}}.(Inverse[aa].h0.h))[[1]];a2=-((v.aa).{{0,0,1/2},{0,0,0},{0,0,0}}.(Inverse[aa].h0.h))[[1]];
If[print,Plot[E^(-t \[Lambda]1) (a0+a1 t+a2 t^2),{t,0,5}];Print["at t=1:",E^(-t \[Lambda]1) (a0+a1 t+a2 t^2)/.t->1];
Print["{\[Lambda]=",\[Lambda]1,", a0=",a0,", a1=",a1,", a2=",a2,"}"];];If[a0<0,Print["negative density at 0!!!"];
Return[res={2,\[Lambda]1,a0,a1,a2,False}];];If[a2<0,Print["negative density at \[Infinity]!!!"];
Return[res={2,\[Lambda]1,a0,a1,a2,False}];];If[a1<-2 Sqrt[a0 a2],Print["negative density on (0,\[Infinity])!!!"];
Return[res={2,\[Lambda]1,a0,a1,a2,False}];];Return[res={2,\[Lambda]1,a0,a1,a2,True}];];
If[\[Lambda][[1]]\[NotElement]Reals,If[print,Print["CASE3: complex eigenvalues"]];
{a1,a2,a3}=-v.aa Transpose[Inverse[aa].h0.h][[1]];{\[Lambda]1,\[Lambda]2,\[Lambda]3}=\[Lambda];
ar=Re[a3];\[Lambda]r=\[Lambda]3;\[Omega]=Abs[Im[\[Lambda]1]];ac=2 Abs[a1];\[Lambda]c=Re[\[Lambda]1];\[Phi]=ArcTan[Re[a1],Abs[Im[a1]]];
If[print,Plot[ar E^(-\[Lambda]r t)+ac E^(-\[Lambda]c t) Cos[\[Omega] t+\[Phi]],{t,0,5}];
Print["at t=1:",ar E^(-\[Lambda]r t)+ac E^(-\[Lambda]c t) Cos[\[Omega] t+\[Phi]]/.t->1];
Print["{ar=",ar,", \[Lambda]r=",\[Lambda]r,", ac=",ac,", \[Lambda]c=",\[Lambda]c,", \[Omega]=",\[Omega],", \[Phi]=",\[Phi],"}"];];
If[ar+ac Cos[\[Phi]]<0,Print["negative density at 0!!!"];Return[res={3,ar,\[Lambda]r,ac,\[Lambda]c,\[Omega],\[Phi],False}];];
If[\[Lambda]r>\[Lambda]c,Print["\[Lambda]r>\[Lambda]c, negative density at 0!!!"];Return[res={3,ar,\[Lambda]r,ac,\[Lambda]c,\[Omega],\[Phi],False}];];
If[ar<0,Print["negative density at \[Infinity]!!!"];Return[res={3,ar,\[Lambda]r,ac,\[Lambda]c,\[Omega],\[Phi],False}];];
If[ac>ar E^(((\[Lambda]c-\[Lambda]r) 2 \[Pi])/\[Omega]),Print["a2 is too large, negative density on (0,2\[Pi]/\[Omega])!!!"];
Return[res={3,ar,\[Lambda]r,ac,\[Lambda]c,\[Omega],\[Phi],False}];];If[ac<ar E^(((\[Lambda]c-\[Lambda]r) 2 \[Pi])/\[Omega])&&ac>ar,tmin=Max[0,(\[Pi]/2-\[Phi])/\[Omega]];
tmax=Min[(\[Pi]-\[Phi])/\[Omega],Log[ac/ar]/(\[Lambda]c-\[Lambda]r)];If[print,Print["(tmin,tmax)=",tmin,", ",tmax]];
ft=ar E^(-\[Lambda]r t)+ac E^(-\[Lambda]c t) Cos[\[Omega] t+\[Phi]];If[(ft/.t->tmax)<0,Print["negative density at tmax!!!"];
Return[res={3,ar,\[Lambda]r,ac,\[Lambda]c,\[Omega],\[Phi],False}];];fvt=\!\(
\*SubscriptBox[\(\[PartialD]\), \(t\)]ft\);If[print,Plot[fvt,{t,tmin,tmax},AxesOrigin->{0,0}]];
If[(fvt/.t->tmin)>0,If[print,Print["increasing density at tmin!!!"]];Return[res={3,ar,\[Lambda]r,ac,\[Lambda]c,\[Omega],\[Phi],True}];];
If[(fvt/.t->tmax)<0,If[print,Print["decreasing density at tmax!!!"]];Return[res={3,ar,\[Lambda]r,ac,\[Lambda]c,\[Omega],\[Phi],True}];];
If[(fvt/.t->tmax)>0,rrr=FindRoot[0==fvt,{t,(tmax+tmin)/2}];If[print,Print["root in tmin,tmax:",rrr]];
If[rrr[[1,2]]>=tmax||rrr[[1,2]]<=tmin,rrr=FindRoot[0==fvt,{t,tmax}];
If[print,Print["root in tmin,tmax:",rrr]];];
If[rrr[[1,2]]>=tmax||rrr[[1,2]]<=tmin,rrr=FindRoot[0==fvt,{t,tmin}];
If[print,Print["root in tmin,tmax:",rrr]];];If[rrr[[1,2]]>=tmax||rrr[[1,2]]<=tmin,res={3,ar,\[Lambda]r,ac,\[Lambda]c,\[Omega],\[Phi],False};
Print["find minimum error!! no root found in (tmin,tmax)!!! res=",res];Return[res={3,ar,\[Lambda]r,ac,\[Lambda]c,\[Omega],\[Phi],False}];];
If[(ft/.rrr)<0,Print["negative density on (tmin,tmax)!!!"];Return[res={3,ar,\[Lambda]r,ac,\[Lambda]c,\[Omega],\[Phi],False}];];];];
Return[res={3,ar,\[Lambda]r,ac,\[Lambda]c,\[Omega],\[Phi],True}];];];



MEContOrder[\[Alpha]_,A_]:=
Module[{size,h,n,res},
size=Dimensions[A][[1]];
h=Table[1,{size}];
res=MatrixRank[Table[h.MatrixPower[Transpose[A],n-1],{n,size}]];
Return[res];
];



MEObsOrder[\[Alpha]_,A_]:=
Module[{size,n,res},
size=Dimensions[A][[1]];
res=MatrixRank[Table[\[Alpha].MatrixPower[A,n-1],{n,size}]];
Return[res];
];



MEMinimize[\[Alpha]in_,Ain_]:= 
Module[ {\[Alpha], A},
{\[Alpha], A}=MEContMinimize[\[Alpha]in,Ain]//Chop ;
{\[Alpha], A}=MEObsMinimize[\[Alpha], A] //Chop ;
Return[{\[Alpha], A}];
]



MEContMinimize[\[Alpha]in_,Ain_]:= 
Module[ {d0,d1,size,vec,B,n,i,rowsum},
size=Dimensions[Ain][[1]];
vec= Table[1,{size},{1}];
d0=Ain; d1=-Ain.vec.{\[Alpha]in};

{n,B}=StairCase1[d0,d1,vec];
d0=Inverse[B].d0.B;
d1=Inverse[B].d1.B;
d0=d0[[1;;n,1;;n]];
d1=d1[[1;;n,1;;n]];

For[i=1,i<=n,i++,
rowsum=d1[[i]].Table[1,{n}];
If[rowsum!=0,
Return[{d1[[i]]/rowsum,d0}]]
];
Return[{d1[[1]],d0}];
]


MEObsMinimize[\[Alpha]in_,Ain_]:= 
Module[ {d0,d1,size,vec,B,n,i,rowsum},
size=Dimensions[Ain][[1]];
vec= Table[1,{size},{1}];
d0=Ain; d1=-Ain.vec.{\[Alpha]in};

vec={\[Alpha]in};
{n,B}=StairCase1[Transpose[d0],Transpose[d1],Transpose[vec]];
d0=Inverse[B].d0.B;
d1=Inverse[B].d1.B;
d0=d0[[1;;n,1;;n]];
d1=d1[[1;;n,1;;n]];

For[i=1,i<=n,i++,
rowsum=d1[[i]].Table[1,{n}];
If[rowsum!=0,
Return[{d1[[i]]/rowsum,d0}]]
];
Return[{d1[[1]],d0}];

]


StairCase1[Xin_,Yin_,Zin_]:=
Module[{\[Epsilon],X,Y,Z,m,U,ranksum,crit,r,Ui,S,T,TEMP,Z1,X1,X2,X3,X4,Y1,Y2,Y3,Y4,Transf,I,n,x,i,nonzero,zeros,R,
y,\[CapitalGamma],TEMP1,sizeUi},
\[Epsilon]=10^-8;  (* numerical precision *)
If[BuToolsVerbose,Print["Numerical precision of StairCase is:", \[Epsilon]//N, " !!" ]; ];
X=Xin;Y=Yin;Z=Zin//N;
m=Dimensions[X][[1]];
U=IdentityMatrix[m];
ranksum=0;
crit=True;
While[crit, 
r=MatrixRank[Z];
ranksum+=r;
{Ui,S,T}=SingularValueDecomposition[Z];
TEMP=Transpose[Ui].Z;
Z1=TEMP[[1;;r,All]];

TEMP=Transpose[Ui].X.Ui;
X1=TEMP[[1;;r,1;;r]];
X2=TEMP[[1;;r,r+1;;-1]];
X3=TEMP[[r+1;;-1,1;;r]];
X4=TEMP[[r+1;;-1,r+1;;-1]];

TEMP=Transpose[Ui].Y.Ui;
Y1=TEMP[[1;;r,1;;r]];
Y2=TEMP[[1;;r,r+1;;-1]];
Y3=TEMP[[r+1;;-1,1;;r]];
Y4=TEMP[[r+1;;-1,r+1;;-1]];

sizeUi=Dimensions[Ui][[1]];
Transf=IdentityMatrix[ranksum-r+sizeUi];
Transf[[-sizeUi ;;-1,-sizeUi ;;-1]]=Transpose[Ui];
U=Transpose[Transf.Transpose[U]];


X=X4;Y=Y4;
Z=Join[X3,Y3,2];

(* Print["Z=",Z]; *)
If[Max[SingularValueList[Z]]<\[Epsilon] ||MatrixRank[Z]==m-ranksum,crit=False];
];

I=IdentityMatrix[m];

If[Max[SingularValueList[Z]]<\[Epsilon],
n=ranksum;
x=Transpose[U].Table[  1,{m},{1}];
x=x[[1;;n,All]];

R=IdentityMatrix[n];
If[Min[Abs[x]]<\[Epsilon],
nonzero=0;
zeros=Table[False,{n}];
For[i=1,i<=n,i++,
If[Abs[x[[i,1]]]<\[Epsilon],zeros[[i]]=True,nonzero=i];
];
(* Print["nonzero=",nonzero, "; zeros=",zeros ]; *)

If[nonzero !=0,
For[i=1,i<=n,i++, If[zeros[[i]],R[[i,nonzero]]=1]
]; 
];
];

y=R.x;
(* Print["x=",x, "; R=",R,"; y=",y]; *)

\[CapitalGamma]=DiagonalMatrix[Flatten[y]];
TEMP=I;
TEMP[[1;;n,1;;n]]=Inverse[\[CapitalGamma]];
TEMP1=I;
TEMP1[[1;;n,1;;n]]=R;

Return[{n,Inverse[TEMP.TEMP1.Transpose[U]]}];
];

If[MatrixRank[Z]==m-ranksum,Return[{m,I}]];
]



MEDensity[\[Alpha]_,A_,x_]:= -\[Alpha].A.MatrixExp[A x].Table[1,{Dimensions[A][[1]]}];



CheckMEPositiveDensity[\[Alpha]_,A_]:=
Module[{\[Beta],B,res},
If[CheckPH[\[Alpha],A],Print["CheckMEPositiveDensity: The input is a PH distribution"];Return[True]];
If[Not[CheckME[\[Alpha],A]],Return[False]]; 
{\[Beta],B}=MonocyclicRepresentation[\[Alpha],A];
If[BuToolsVerbose,
Print["size=",Dimensions[B][[1]], ",  \[Beta]=",\[Beta],", B=",MatrixForm[B]];
];
res=PositiveRep[\[Beta],B];
Return[res];
]



(* Check positive representation 
       with density check at maximal negative exit probability  and
       with initial vector shifting
Input: monocyclic representation 
Output: true - false *)
PositiveRep[\[Alpha]in_,A_]:=
Module[{\[Epsilon],\[Alpha],\[Lambda],\[Lambda]dominant, size,vec,out,A\[Lambda], h,i,iter,niter,poziter,outold,shift,\[CapitalDelta]shift},
\[Epsilon]=10^-10//N;
\[Alpha]=\[Alpha]in;
size=Dimensions[\[Alpha]][[1]];
h=Table[1,{size}];
shift=0;

If[-\[Alpha].A.h<-\[Epsilon],Print["negativ density at 0 (at precision:",\[Epsilon],")"];Return[False];];
If[(\[Alpha].A.h==0 && -\[Alpha].A.A.h<-\[Epsilon]),Print["0 density at 0 and negative derivative (at precision:",\[Epsilon],")"];Return[False];];
 
\[Lambda]dominant=Re[Eigenvalues[-A]];
\[Lambda]dominant=Max[\[Lambda]dominant//N];
\[Lambda]=\[Lambda]dominant/2;
niter=0;
iter=0;

For[i=1,i<18,i++,
niter+=iter;
iter=0;
\[Lambda]=2\[Lambda];
vec=\[Alpha];
A\[Lambda]=(IdentityMatrix[size]+A/\[Lambda]);
out= vec.(h - A\[Lambda].h); 

While[Min[vec]<0,
iter++;
vec= vec . A\[Lambda];
out= vec.(h - A\[Lambda]. h); 
(*  Print["!! vec=",vec, ",  out=",out,", \[Lambda]=",\[Lambda], ", i=", i, ", iter=",iter]; *)
If[out<0,iter--;Break[]];
];

(* ha megvan a pozotiv reprezentacio visszaterek *)
If[Min[vec]>0,
niter+=iter;
If[BuToolsVerbose,
Print["vec=",vec, ",  out=",out,", \[Lambda]=",\[Lambda], ", i=", i, ", iter=",iter, ", niter=",niter];
];
Return[True]];

(* ha nem megkeresem a negativ szakasz aljat ... *)
If[out<0,
outold=0;
poziter=iter;
While[out< outold,
iter++;
outold=out;
vec= vec . A\[Lambda];
out= vec.(h - A\[Lambda]. h); 
(* Print["vec=",vec, ",  out=",out,", \[Lambda]=",\[Lambda], ", i=", i, ", iter=",iter]; *)
];

(* ... es megnezem, hogy ott negativ-e a suruseg *)
(* Print["Density at iter/\[Lambda]:",-\[Alpha].MatrixExp[A iter/\[Lambda]].A.h]; *)
If[-\[Alpha].MatrixExp[A iter/\[Lambda]].A.h<0,
niter+=iter;
If[BuToolsVerbose,
Print["Negative density at ",shift+iter/\[Lambda], ": f(x)=",-\[Alpha].MatrixExp[A iter/\[Lambda]].A.h,"!!!  \[Lambda]=",\[Lambda], ", i=", i, ", iter=",iter, ", niter=",niter]; 
];
If[(shift+iter/\[Lambda])  \[Lambda]dominant >20,Print["Time * \[Lambda]dominant=",(shift+iter/\[Lambda])  \[Lambda]dominant," > 20 !!!!  It might be a numerical error!!!"]];
(* If[Abs[-\[Alpha].MatrixExp[A iter/\[Lambda]].A.h]<10^-20,Print["f[x]<10^-20, it might be a numerical error!!!"]]; *)
Return[False];
];
];

(* ... ha nem eltolom a kezdo vektort ... *)
(* Print["Density at iter/\[Lambda]:",-\[Alpha].MatrixExp[A iter/\[Lambda]].A.h]; *)

\[CapitalDelta]shift=iter/\[Lambda]; (* ... a negativ szakasz aljara *)
\[CapitalDelta]shift=poziter/\[Lambda]; (* ... a pozitiv szakasz vegere *)

\[Alpha] =\[Alpha].MatrixExp[A  \[CapitalDelta]shift];
If[BuToolsVerbose,
Print["\[CapitalDelta] shift=", \[CapitalDelta]shift,", time shift=",shift+=\[CapitalDelta]shift,", vec=",\[Alpha]];
];

];
Print["Could not decide the non-negativity till \[Lambda]=",\[Lambda], ", i=", i, ", niter=",niter]; 
Return[False];
];



End[(* Private *)]

Protect[MomentsFromME, MomentsFromPH, MEFromMoments, Appie, 
PH3Canonical, ME3Canonical,
METoPH, PH2Canonical, APHRepresentation, MonocyclicRepresentation, 
APHFrom3Moments, MEMoments, MEOrderFromMoments, ME3member,
MEContOrder, MEObsOrder, MEObsMinimize, MEContMinimize, MEMinimize, MEDensity, CheckMEPositiveDensity];

EndPackage[]
