(* ::Package:: *)

(*
   BuTools_SpecialProcesses Package
*)


BeginPackage["BuTools`SpecialProcesses`"];

MStairCase::usage = "MStairCase [ vector of matrix0, matrix1, ... matrixK, closing vector ] -> [ size, similarity matrix ] : Computes a smaller representation of an MRAP using staircase algorithm."
MRAPObsMinimize::usage = "MRAPObsMinimize [ vector of matrix0, matrix1, ... matrixK ] -> [ vector of matrix0, matrix1, ... matrixK ] : Computes a minimal representation of an MRAP using the staircase methor by eliminating the redundancy caused by the initial vector."
MRAPContMinimize::usage = "MRAPContMinimize [ vector of matrix0, matrix1, ... matrixK ] -> [ vector of matrix0, matrix1, ... matrixK ] : Computes a minimal representation of an MRAP using the staircase methor by eliminating the redundancy caused by the closing vector."
MRAPMinimize::usage = "MRAPMinimize [ vector of matrix0, matrix1, ... matrixK ] -> [ vector of matrix0, matrix1, ... matrixK ] : Computes a minimal representation of an MRAP using the staircase method by eliminating the redundancy caused by the initial vector and by the closing vector."
MomentsFromMRAP::usage = "MomentsFromMRAP [ vector of matrix0, matrix1, ... matrixK ] -> [ moments ] : Calculates the first 2n-1 marginal moments of the MRAP of size n."
Lag1JointMomentsFromMRAP::usage = "Lag1JointMomentsFromMRAP [ vector of matrix0, matrix1, ... matrixK] -> [ vector of jointmomentsmatrix1, ... jointmomentsmatrixK ] : calculates the defective joint moments of (X0^(k))^i X1^j (type k) for i,j=0 to n-1, k=1 to K and returns the result in vector of n x n  matrices."
MRAPFromMoments::usage = "MRAPFromMoments [ marginal moments, vector of jointmomentsmatrix1, ... jointmomentsmatrixK ] -> [ vector of matrix0, matrix1, ... matrixK ] : Calculates an MRAP representation based on 2n-1 marginal moments and the vector of nxn matrices of the lag1 joint moments."
MRAPToMMAP::usage = "MRAPToMMAP [  vector of matrix0, matrix1, ... matrixK ] -> [  vector of matrix0, matrix1, ... matrixK ] : Similarity transforms the non-Markovian representation of a marked rational arrival process to the Markovian representation of the same size if possibe"

TRAPObsMinimize::usage = "TRAPObsMinimize [  vector, matrix0, matrix1 ] -> [  vector, matrix0, matrix1 ] : Computes a minimal representation of an TRAP using the staircase methor by eliminating the redundancy caused by the initial vector."
TRAPContMinimize::usage = "TRAPContMinimize [  vector, matrix0, matrix1 ] -> [  vector, matrix0, matrix1 ] : Computes a minimal representation of an TRAP using the staircase methor by eliminating the redundancy caused by the closing vector."
TRAPMinimize::usage = "TRAPMinimize [  vector, matrix0, matrix1 ] -> [  vector, matrix0, matrix1 ] : Computes a minimal representation of an TRAP using the staircase method by eliminating the redundancy caused by the initial vector and by the closing vector."
MomentsFromTRAP::usage = "MomentsFromTRAP [ vector (1 x n), matrix0 (n x n), matrix1 (n x n) ] -> [ moments ] :  returns: defective moments of X0^i for i=0 to 2n-1"
LagkJointMomentsFromTRAP::usage = "LagkJointMomentsFromTRAP [ vector (1 x n), matrix0 (n x n), matrix1 (n x n^2), k ] -> [ moments ] : returns: defective joint moments of X0^i Xk^j for i,j= 0 to n-1"
TRAPFromMoments::usage = "TRAPFromMoments [  vector (1 x 2n), matrix0 (n x n) ] -> [ vector (1 x n), matrix0 (n x n), matrix1 (n x n) ] :  input: X0^i defective moments for i=0 to 2n-1 + defective joint moments of X0^i X1^j for i,j=0 to n-1. It returns a representation of the TMAP."
TRAPToTMAP::usage = "TRAPToTMAP [ vector, matrix0, matrix1 ] -> [ vector, matrix0, matrix1 ] : Similarity transforms the non-Markovian representation of a transient rational arrival process to the Markovian representation of the same size if possibe"

RBTObsMinimize::usage = "RBTObsMinimize [  vector (1 x n), matrix0 (n x n), matrix1 (n x n^2) ] -> [  vector (1 x n), matrix0 (n x n), matrix1 (n x n^2) ] : Computes a minimal representation of an RBT using the staircase methor by eliminating the redundancy caused by the initial vector."
RBTContMinimize::usage = "RBTContMinimize [  vector (1 x n), matrix0 (n x n), matrix1 (n x n^2) ] -> [  vector (1 x n), matrix0 (n x n), matrix1 (n x n^2) ] : Computes a minimal representation of an RBT using the staircase methor by eliminating the redundancy caused by the closing vector."
RBTMinimize::usage = "RBTMinimize [  vector (1 x n), matrix0 (n x n), matrix1 (n x n^2) ] -> [  vector (1 x n), matrix0 (n x n), matrix1 (n x n^2) ] : Computes a minimal representation of an RBT using the staircase method by eliminating the redundancy caused by the initial vector and by the closing vector."
GammaijkMomentsFromRBT::usage = "GammaijkMomentsFromRBT [ vector (1 x n), matrix0 (n x n), matrix1 (n x n^2) ] -> [ moments ] : calculates the defective joint moments of X0^i X1^j Y0^k for i,j,k=0 to n-1 and returns them in an n x n^2 matrix."
MomentsFromRBT::usage = "MomentsFromRBT [ vector (1 x n), matrix0 (n x n), matrix1 (n x n^2) ] -> [ moments ] : returns: defective moments of X0^i for i=0 to 2n-1."
RBTFromMoments::usage = "RBTFromMoments [  vector (1 x 2n), matrix0 (n x n^2) ] -> [ vector (1 x n), matrix0 (n x n), matrix1 (n x n^2) ] : input: X0^i defective moments for i=0 to 2n-1 + defective joint moments of X0^i X1^j Y0^k for i,j,k=0 to n-1. It returns a representation of the MBT. "
RBTToMBT::usage = "RBTToTMAP [ vector, matrix0, matrix1 ] -> [ vector, matrix0, matrix1 ] : Similarity transforms the non-Markovian representation of a rational binary tree process to a Markovian representation of the same size if possible."


Unprotect[RBTObsMinimize, RBTContMinimize, RBTMinimize, GammaijkMomentsFromRBT, MomentsFromRBT, RBTFromMoments, 
TRAPObsMinimize, TRAPContMinimize, TRAPMinimize, MomentsFromTRAP, LagkJointMomentsFromTRAP, TRAPFromMoments,
MStairCase, MRAPObsMinimize, MRAPContMinimize, MRAPMinimize, MomentsFromMRAP, Lag1JointMomentsFromMRAP, MRAPFromMoments, 
MRAPToMMAP, TRAPToTMAP, RBTToMBT ];


Begin["`Private`"]

Needs["BuTools`Utilities`"]
Needs["BuTools`PH`"]
Needs["BuTools`MAP`"]


RBTMinimize[\[Alpha]_,D0_,B_]:= 
Module[ {\[Gamma],G0,H},
{\[Gamma],G0,H}=RBTContMinimize[\[Alpha],D0,B]//Chop ;
{\[Gamma],G0,H}=RBTObsMinimize[\[Gamma],G0,H]//Chop ;
Return[{\[Gamma],G0,H}];
]


RBTContMinimize[\[Alpha]_,D0_,B_]:= 
Module[ {\[Gamma],G0,H,H3D,i,j,k,size, vec,SS ,n},
(* Perform consistency check !! *)
size=Dimensions[D0][[1]];
vec= Table[1,{size},{1}];
{n,SS}=StairCaseRBTCont[D0,B,vec];

\[Gamma]=(\[Alpha].SS)[[1;;n]];
G0=(Inverse[SS].D0.SS)[[1;;n,1;;n ]];
H=Inverse[SS].B.KroneckerProduct[SS,SS];
H3D =Table[ H[[i,(j-1) size + k]],{i,size},{j,size},{k,size}];
H3D = H3D [[1;;n,1;;n ,1;;n ]] ;
H =Table[ H3D[[i,Floor[(j-1)/n]+1,Mod[j-1,n]+1 ]],{i,n},{j,n^2}];

Return[{\[Gamma],G0,H}];
]


StairCaseRBTCont[D0in_,Bin_,Zin_]:=
Module[{\[Epsilon],X,Y,Z,m,U,ranksum,crit,r,Ui,S,T,TEMP,TEMP3D,X1,X2,X3,X4,
Y1,Y2,Y3,Y4,Y5,Y6,Y7,Y8,Y13D,Y23D,Y33D,Y43D,Y53D,Y63D,Y73D,Y83D,
Transf,I,n,x,i,nonzero,zeros,R,
y,\[CapitalGamma],TEMP1,sizeUi,mm,ii},
\[Epsilon]=10^-8;  (* numerical precision *)
X=D0in;
Y=Bin;
Z=Zin//N;
m=Dimensions[X][[2]];
U=IdentityMatrix[m];
ranksum=0;
crit=True;
While[crit, 
r=MatrixRank[Z];
ranksum+=r;
{Ui,S,T}=SingularValueDecomposition[Z];
TEMP=Transpose[Ui].Z;

(* Print["Ui=",Ui//MatrixForm,", TempZ=",TEMP, ", r=",r,", ranksum=", ranksum]; *)
(* Z1=TEMP[[1;;r,All]]; *)

mm=Dimensions[X][[1]];
TEMP=Transpose[Ui].X.Ui;
X1=TEMP[[1;;r,1;;r]];
X2=TEMP[[1;;r,r+1;;mm]];
X3=TEMP[[r+1;;mm,1;;r]];
X4=TEMP[[r+1;;mm,r+1;;mm]];
(* Print["TempX=",TEMP//MatrixForm, ", X3=",X3,", X4=", X4]; *)

TEMP=Transpose[Ui].Y.KroneckerProduct[Ui,Ui];
TEMP3D =Table[ TEMP[[i,(j-1) mm + k]],{i,mm},{j,mm},{k,mm}];
(*
Y13D=TEMP3D[[1;;r,1;;r,1;;r]];
Y23D=TEMP3D[[1;;r,1;;r,r+1;;mm]]; *)
Y33D=TEMP3D[[r+1;;mm,1;;r,1;;r]];
(* Y43D=TEMP3D[[r+1;;mm,1;;r,r+1;;mm]];
Y53D=TEMP3D[[1;;r,r+1;;mm,1;;r]];
Y63D=TEMP3D[[1;;r,r+1;;mm,r+1;;mm]];
Y73D=TEMP3D[[r+1;;mm,r+1;;mm,1;;r]]; *)
Y83D=TEMP3D[[r+1;;mm,r+1;;mm,r+1;;mm]];

Y3 =Table[ Y33D[[i,Floor[(j-1)/r]+1,Mod[j-1,r]+1 ]],{i,(mm-r)},{j,r^2}];
(* Y4 =Table[ Y43D[[i,Floor[(j-1)/(mm-r)]+1,Mod[j-1,(mm-r)]+1 ]],{i,(mm-r)},{j,r (mm-r)}];
Y7 =Table[ Y73D[[i,Floor[(j-1)/r]+1,Mod[j-1,r]+1 ]],{i,(mm-r)},{j, (mm-r) r}]; *)
Y8 =Table[ Y83D[[i,Floor[(j-1)/(mm-r)]+1,Mod[j-1,(mm-r)]+1 ]],{i,(mm-r)},{j,(mm-r)^2}];
(* Print["TempY=",TEMP//MatrixForm, ", Y3=",Y3,", Y4=",Y4,", Y7=",Y7,", Y8=", Y8]; *)

sizeUi=Dimensions[Ui][[1]];
Transf=IdentityMatrix[ranksum-r+sizeUi];
Transf[[-sizeUi ;;-1,-sizeUi ;;-1]]=Transpose[Ui];
U=Transpose[Transf.Transpose[U]];

X=X4;
Y=Y8;
Z={{}};
Z=Join[Z,X3,2];
Z=Join[Z,Y3,2];
(* Z=Join[Z,Y4,2]; *)
(* Z=Join[Z,Y7,2]; *)

(* Print["Z=",Z//MatrixForm, ", rank Z=",MatrixRank[Z], ", Z=0:",(Max[SingularValueList[Z]]<\[Epsilon]), ", Z is full rank: ",(MatrixRank[Z]==m-ranksum)]; 
Print["SingularValueList[Z]=",SingularValueList[Z]];  *)
If[Max[SingularValueList[Z]]<\[Epsilon] ||MatrixRank[Z]==m-ranksum,crit=False];
];

I=IdentityMatrix[m];

If[Max[SingularValueList[Z]]<\[Epsilon],
n=ranksum;
x=Transpose[U].Table[  1,{m},{1}];
x=x[[1;;n,All]];

R=IdentityMatrix[n];
If[Min[Abs[x]]<\[Epsilon],
nonzero=0;
zeros=Table[False,{n}];
For[i=1,i<=n,i++,
If[Abs[x[[i,1]]]<\[Epsilon],zeros[[i]]=True,nonzero=i];
];
(* Print["nonzero=",nonzero, "; zeros=",zeros ]; *)
If[nonzero !=0,
For[i=1,i<=n,i++, If[zeros[[i]],R[[i,nonzero]]=1]
]; 
];
];

y=R.x;
(* Print["x=",x, "; R=",R,"; y=",y]; *)

\[CapitalGamma]=DiagonalMatrix[Flatten[y]];
TEMP=I;
TEMP[[1;;n,1;;n]]=Inverse[\[CapitalGamma]];
TEMP1=I;
TEMP1[[1;;n,1;;n]]=R;

Return[{n,Inverse[TEMP.TEMP1.Transpose[U]]}];

];

If[MatrixRank[Z]==m-ranksum,Return[{m,I}]];
];


RBTObsMinimize[\[Alpha]_,D0_,B_]:= 
Module[ {\[Gamma],G0,H,XMX,H3D,i,j,k,size, vec,SS ,n},
(* Perform consistency check !! *)
size=Dimensions[D0][[1]];
XMX={Transpose[D0]};
For [i=1,i<=size,i++,
XMX=Append[XMX,Transpose[B[[All, (i-1)size +1;;i  size]]] ];
];
{n,SS}=MStairCase[XMX,Transpose[{\[Alpha]}]];

\[Gamma]=(\[Alpha].SS)[[1;;n]];
G0=(Inverse[SS].D0.SS)[[1;;n,1;;n ]];
H=Inverse[SS].B.KroneckerProduct[SS,SS];
H3D =Table[ H[[i,(j-1) size + k]],{i,size},{j,size},{k,size}];
H3D = H3D [[1;;n,1;;n ,1;;n ]] ;
H =Table[ H3D[[i,Floor[(j-1)/n]+1,Mod[j-1,n]+1 ]],{i,n},{j,n^2}];

Return[{\[Gamma],G0,H}];
]


RBTFromMoments[mom_,nn_]:=
(* input: X0 defective moments from 0 to 2n-1  *)
(* + defective joint moments of X0 X1 Y0 from 0 to n-1  *)
Module[{mom1,\[Mu]0, Rhat,size,hh,gammav,gamma1,\[Phi],h0,h0i,i,j,k,l,vv,vvv, w,res,\[Epsilon],nonzero,zeros,B,Bi},
(* one might check the sizes of the input matrices !!!!!  *)
\[Mu]0=mom[[1]];
mom1=Drop[mom,1]/\[Mu]0;
{\[Phi],h0}=MEFromMoments[mom1];
(* Print["\[Phi]=",\[Phi], "; h0=",h0 ]; *)

h0i=Inverse[-h0];
size=Dimensions[\[Phi]][[1]];
hh=Table[1,{size}];

gammav=Table[0,{size},{size}];
For[i=1,i<=size,i++,gammav[[i]]=(i-1)!  \[Phi].MatrixPower[h0i,i-1];];
gamma1=Table[0,{size},{size},{size^2}];
For[i=1,i<=size,i++,
For[j=1,j<=size,j++,
(* gamma1[[i,j]]= Flatten[Table[Subscript[a, i,j,k,l],{k,2},{l,2}]]; *)
 gamma1[[i,j]]=Flatten[KroneckerProduct[
\[Mu]0 (i-1)!  MatrixPower[h0i,i-1].hh,\[Mu]0 (j-1)!  MatrixPower[h0i,j-1].hh]]; 
 (* Print["i,j:",i,j," gamma:",gamma1[[i,j]]]; *)
];
];
(* Print["gamma1 aa:",gamma1]; *)
gamma1=Flatten[gamma1,1];
(* Print["gamma1 bb:",gamma1]; *)
gamma1=Transpose[gamma1];
(*  Print["gamma1 cc:",gamma1]; *)

(*  Print[gammav//MatrixForm,gamma1//MatrixForm]; *)
Rhat=Inverse[gammav].nn.Inverse[gamma1];
 (* Return[{\[Phi],h0,-h0.Rhat,\[Mu]0 hh}];  !!!!!!!!!!!!!!!!!!!!!! *)

(* now we have \[Phi],h0,Rhat, we still need the realted closing vector vv 
satisfying \[Mu]0 hh == Rhat . KroneckerProduct[vv,vv] *)
\[Epsilon]=10^-6;
vv=Table[Subscript[w, i],{i,size}];
res=Solve[Rhat.Flatten[ KroneckerProduct[vv,vv]]==\[Mu]0 hh,vv//Flatten];
(* Print["res=",res ];  *)
For[i=1,i<=Dimensions[res][[1]],i++,
vvv=vv/.res[[i]];
(* Print["\[Phi].vv=",\[Phi].vvv];*)
If[Abs[\[Phi].vvv-1]<\[Epsilon],Break[]];
];
vv=vv/.res[[i]];

(* Print["Rhat=",Rhat, "; vv=",vv ];  *)

(* now we have to transform \[Phi],h0,Phat,vv such that the closing vector becomes 1 *)
nonzero=0;
zeros=Table[False,{size}];
For[i=1,i<=size,i++,
If[Abs[vv[[i]]]<\[Epsilon],zeros[[i]]=True,nonzero=i];
];
(*  Print["nonzero=",nonzero, "; zeros=",zeros ]; *)

B=DiagonalMatrix[vv];
If[nonzero !=0,
For[i=1,i<=size,i++, 
If[zeros[[i]],
B[[i,i]]=1;B[[i,nonzero]]=-1
]
]; 
];
If[\[Phi].B.hh<0,B=-B];
Bi=Inverse[B];

(* return: \[Alpha], d0 and b and closing *)
{\[Phi],h0,Rhat}={\[Phi].B,Bi.h0.B,Bi.Rhat.KroneckerProduct[B,B]};
Return[{\[Phi],h0,-h0.Rhat}];
] 


MomentsFromRBT[\[Alpha]_,d0_,b_]:=
(* returns: defective moments of X0 from 0 to 2n-1  *)
Module[{size,moms,i},
size= Dimensions[d0][[1]];
moms=Table[i!  \[Alpha].MatrixPower[Inverse[-d0],i+1].b.Table[1,{size^2}],{i,0,2 size -1}];
Return[moms];
];


GammaijkMomentsFromRBT[\[Alpha]_,d0_,b_]:=
(* returns: defective joint moments of X0 X1 Y0 from 0 to n-1 
in an n x n^2 matrix *)
Module[{h,hh,n,d0i,res,i,j,k,R},
n = Dimensions[d0][[1]];
h=Table[1,{n}];
hh=Table[1,{n^2}];
d0i=Inverse[-d0];
R=d0i.b;
res=Table[  i! j!k!\[Alpha].MatrixPower[d0i,i].R.
Flatten[KroneckerProduct[MatrixPower[d0i,j].R.hh,MatrixPower[d0i,k].R.hh]],
{i,0,n-1},{j,0,n-1}, {k,0,n-1}];
Return[ Table[ Flatten[res[[i]]],{i,n}] ];
]



TRAPMinimize[\[Alpha]_,D0_,D1_]:= 
Module[ {\[Gamma],G0,G1},
{\[Gamma],G0,G1}=TRAPContMinimize[\[Alpha],D0,D1]//Chop ;
{\[Gamma],G0,G1}=TRAPObsMinimize[\[Gamma],G0,G1]//Chop ;
Return[{\[Gamma],G0,G1}];
]


TRAPContMinimize[\[Alpha]_,D0_,D1_]:= 
Module[ {XMX,\[Gamma],PP,G0,G1,ii,size, vec,B ,n},
XMX={D0,D1};
PP=Inverse[-D0].D1;
size=Dimensions[D0][[1]];
vec= Table[1,{size},{1}];
(* {n,B}=MStairCase[XMX,PP.vec]; *)
{n,B}=MStairCase[XMX,vec];

\[Gamma]=(\[Alpha].B)[[1;;n]];
G0=(Inverse[B].D0.B)[[1;;n,1;;n]];
G1=(Inverse[B].D1.B)[[1;;n,1;;n]];

Return[{\[Gamma],G0,G1}];
]


TRAPObsMinimize[\[Alpha]_,D0_,D1_]:= 
Module[ {XMX,\[Gamma],PP,G0,G1,ii,size, vec,B ,n},
XMX={Transpose[D0],Transpose[D1]};
{n,B}=MStairCase[XMX,Transpose[{\[Alpha]}]];

\[Gamma]=(\[Alpha].B)[[1;;n]];
G0=(Inverse[B].D0.B)[[1;;n,1;;n]];
G1=(Inverse[B].D1.B)[[1;;n,1;;n]];

Return[{\[Gamma],G0,G1}];
]


TRAPFromMoments[mom_,nn_]:=
(* input: X0 defective moments from 0 to 2n-1  *)
(* + defective joint moments of X0 X1 from 0 to n-1  *)
Module[{mom1,\[Mu]0, Phat,size,hh,gammav,gamma1,\[Phi],h0,h0i,i,j,vv,\[Epsilon],nonzero,zeros,B,Bi},
(* one might check the sizes of the input matrices !!!!!  *)
\[Mu]0=mom[[1]];
mom1=Drop[mom,1]/\[Mu]0;
{\[Phi],h0}=MEFromMoments[mom1];
(* Print["\[Phi]=",\[Phi], "; h0=",h0 ]; *)

h0i=Inverse[-h0];
size=Dimensions[\[Phi]][[1]];
hh=Table[1,{size}];

gammav=Table[0,{size},{size}];
For[i=1,i<=size,i++,gammav[[i]]=(i-1)!  \[Phi].MatrixPower[h0i,i-1];];
gamma1=Table[0,{size},{size}];
For[i=1,i<=size,i++,gamma1[[i]]=\[Mu]0 (i-1)!  MatrixPower[h0i,i-1].hh];
gamma1=Transpose[gamma1];

(* Print[gammav//MatrixForm,gamma1//MatrixForm]; *)
Phat=Inverse[gammav].nn.Inverse[gamma1];

(* now we have \[Phi],h0,Phat, we still need the realted closing vector vv *)
vv=\[Mu]0 PseudoInverse[Phat].Transpose[{hh}];
(*Print["Phat=",Phat, "; vv=",vv ]; *)

(* now we have to transform \[Phi],h0,Phat,vv such that the closing vector becomes 1 *)
\[Epsilon]=10^-12;
nonzero=0;
zeros=Table[False,{size}];
For[i=1,i<=size,i++,
If[Abs[vv[[i,1]]]<\[Epsilon],zeros[[i]]=True,nonzero=i];
];
(* Print["nonzero=",nonzero, "; zeros=",zeros ]; *)

B=DiagonalMatrix[Transpose[vv][[1]]];
If[nonzero !=0,
For[i=1,i<=size,i++, 
If[zeros[[i]],
B[[i,i]]=1;B[[i,nonzero]]=-1
]
]; 
];
Bi=Inverse[B];

(* return: \[Alpha], d0 and d1 *)
Return[{\[Phi].B,Bi.h0.B,-Bi.h0.Phat.B}];
] 



MomentsFromTRAP[\[Alpha]_,d0_,d1_]:=
(* returns: defective moments of X0 from 0 to 2n-1  *)
Module[{size,moms,i},
size= Dimensions[d0][[1]];
moms=Table[i!  \[Alpha].MatrixPower[Inverse[-d0],i+1].d1.Table[1,{size}],{i,0,2 size -1}];
Return[moms];
];



LagkJointMomentsFromTRAP[\[Alpha]_,d0_,d1_,k_]:=
(* returns: defective joint moments of X0 X1 from 0 to n-1  *)
Module[{h,n,d0i,res,i,j,P,Pk},
n = Dimensions[d1][[1]];
h=Table[1,{n}];
d0i=Inverse[-d0];
P=d0i.d1;
Pk=MatrixPower[P,k];
res=Table[  i! j!\[Alpha].MatrixPower[d0i,i].Pk.MatrixPower[d0i,j].P.h,{i,0,n-1},{j,0,n-1}];
Return[res];
]



MRAPMinimize[Xin_]:= 
Module[ {X},
X=Xin;
X=MRAPContMinimize[X]//Chop ;
X=MRAPObsMinimize[X] //Chop ;
Return[X];
]


MRAPContMinimize[Xin_]:= 
Module[ {X,Y,msize,ii,size, vec,B ,n},
X=Xin;
msize=Dimensions[X][[1]];
size=Dimensions[X[[1]]][[1]];
vec= Table[1,{size},{1}];
{n,B}=MStairCase[X,vec];
Y=Table[0,{msize},{n},{n}];
For[ii=1,ii<=msize,ii++,
Y[[ii]]=(Inverse[B].X[[ii]].B)[[1;;n,1;;n]];
];
Return[Y];
]




MRAPObsMinimize[Xin_ ]:= 
Module[ {X,XX,Y,d0,di,ii,msize,size, vec,B ,n},
X=Xin;XX=Xin;
msize=Dimensions[X][[1]];
size=Dimensions[X[[1]]][[1]];
d0=X[[1]];
di=Table[0,{size},{size}];
For[ii=2,ii<=msize,ii++,
di+=X[[ii]];
X[[ii]]=Transpose[X[[ii]]];
];
X[[1]]=Transpose[X[[1]]];
(* Print["d0:",d0, ", d1:",d1,", P=",Inverse[-d0].di];*)
vec={DTMCSolve[Inverse[-d0].di]};
{n,B}=MStairCase[X,Transpose[vec]];
(* Print["vec:",vec, ", n:",n,", B=",B]; *)
Y=Table[0,{msize},{n},{n}];
For[ii=1,ii<=msize,ii++,
Y[[ii]]=(Inverse[B].XX[[ii]].B)[[1;;n,1;;n]];
];
Return[Y];
]


MStairCase[Xin_,Zin_]:=
Module[{\[Epsilon],X,Z,msize,m,U,ranksum,crit,r,Ui,S,T,TEMP,Z1,X1,X2,X3,X4,Transf,I,n,x,i,nonzero,zeros,R,
y,\[CapitalGamma],TEMP1,sizeUi,mm,ii},
\[Epsilon]=10^-8;  (* numerical precision *)
X=Xin;Z=Zin//N;
msize=Dimensions[X][[1]];
m=Dimensions[X][[2]];
U=IdentityMatrix[m];
ranksum=0;
crit=True;
While[crit, 
r=MatrixRank[Z];
ranksum+=r;
{Ui,S,T}=SingularValueDecomposition[Z];
TEMP=Transpose[Ui].Z;
Z1=TEMP[[1;;r,All]];

mm=Dimensions[X[[1]]][[1]];
TEMP=Table[0,{msize},{mm},{mm}];
X1=Table[0,{msize},{r},{r}];
X2=Table[0,{msize},{r},{mm-r}];
X3=Table[0,{msize},{mm-r},{r}];
X4=Table[0,{msize},{mm-r},{mm-r}];

For[ii=1,ii<=msize,ii++,
TEMP[[ii]]=Transpose[Ui].X[[ii]].Ui;
X1[[ii]]=TEMP[[ii]][[1;;r,1;;r]];
X2[[ii]]=TEMP[[ii]][[1;;r,r+1;;-1]];
X3[[ii]]=TEMP[[ii]][[r+1;;-1,1;;r]];
X4[[ii]]=TEMP[[ii]][[r+1;;-1,r+1;;-1]];
];

sizeUi=Dimensions[Ui][[1]];
Transf=IdentityMatrix[ranksum-r+sizeUi];
Transf[[-sizeUi ;;-1,-sizeUi ;;-1]]=Transpose[Ui];
U=Transpose[Transf.Transpose[U]];

Z={{}};
mm=Dimensions[X4[[1]]][[1]];
X=Table[0,{msize},{mm},{mm}];

For[ii=1,ii<=msize,ii++,
X[[ii]]=X4[[ii]];
Z=Join[Z,X3[[ii]],2];
];

(* Print["Z=",Z]; *)
If[Max[SingularValueList[Z]]<\[Epsilon] ||MatrixRank[Z]==m-ranksum,crit=False];
];

I=IdentityMatrix[m];

If[Max[SingularValueList[Z]]<\[Epsilon],
n=ranksum;
x=Transpose[U].Table[  1,{m},{1}];
x=x[[1;;n,All]];

R=IdentityMatrix[n];
If[Min[Abs[x]]<\[Epsilon],
nonzero=0;
zeros=Table[False,{n}];
For[i=1,i<=n,i++,
If[Abs[x[[i,1]]]<\[Epsilon],zeros[[i]]=True,nonzero=i];
];
(* Print["nonzero=",nonzero, "; zeros=",zeros ]; *)

If[nonzero !=0,
For[i=1,i<=n,i++, If[zeros[[i]],R[[i,nonzero]]=1]
]; 
];
];

y=R.x;
(* Print["x=",x, "; R=",R,"; y=",y]; *)

\[CapitalGamma]=DiagonalMatrix[Flatten[y]];
TEMP=I;
TEMP[[1;;n,1;;n]]=Inverse[\[CapitalGamma]];
TEMP1=I;
TEMP1[[1;;n,1;;n]]=R;

Return[{n,Inverse[TEMP.TEMP1.Transpose[U]]}];
];

If[MatrixRank[Z]==m-ranksum,Return[{m,I}]];
];



MomentsFromMRAP[Xin_ ]:= 
Module[ {X,d0,di,ii,msize,size},
X=Xin;
msize=Dimensions[X][[1]];
size=Dimensions[X[[1]]][[1]];
d0=X[[1]];
di=Table[0,{size},{size}];
For[ii=2,ii<=msize,ii++,
di+=X[[ii]];
];
(* Print["d0:",d0, ", di:",di,", P=",Inverse[-d0].di];*)
Return[MarginalMomentsFromRAP[d0,di,2size-1]];
]


Lag1JointMomentsFromMRAP[XX_ ]:=
Module[ {d0,d0i,di,i,j,ii,msize,size,vec,hh,JMOM},
msize=Dimensions[XX][[1]];
size=Dimensions[XX[[1]]][[1]];
hh=Table[1,{size}];
d0=XX[[1]];
d0i=Inverse[-d0];
di=Table[0,{size},{size}];
For[ii=2,ii<=msize,ii++,
di+=XX[[ii]];
];
vec=DRPSolve[d0i.di];
 (* Print["d0:",d0, ", di:",di,", vec=",vec]; *)

JMOM=Table[0,{msize-1},{size},{size}];
For[ii=2,ii<=msize,ii++,
di=XX[[ii]];
For[i=1,i<=size,i++,
For[j=1,j<=size,j++,
JMOM[[ii-1,i,j]]=  (i-1)!(j-1)!(vec.MatrixPower[d0i,i-1].d0i.di.MatrixPower[d0i,j-1].hh);
];
];
];
Return[JMOM];
] 


MRAPFromMoments[moms_,JMOM_]:=
Module[ {h0,h0i,gammav, gamma1,i,j,ii,msize,size,vec,hh,XX,rowsum},
msize=Dimensions[JMOM][[1]]+1;
size=Dimensions[JMOM[[1]]][[1]];
hh=Table[1,{size}];

{vec,h0}=MEFromMoments[moms];
h0i=Inverse[-h0];
hh=Table[1,{size}];

gammav=Table[0,{size},{size}];
For[i=1,i<=size,i++,gammav[[i]]=(i-1)!  vec.MatrixPower[h0i,i-1];];
gamma1=Table[0,{size},{size}];
For[i=1,i<=size,i++,gamma1[[i]]=(i-1)!  MatrixPower[h0i,i-1].hh];
gamma1=Transpose[gamma1];

XX=Table[0,{msize},{size},{size}];
XX[[1]]=h0;
For[ii=2,ii<=msize,ii++,
XX[[ii]]=-h0.Inverse[gammav].JMOM[[ii-1]].Inverse[gamma1];
];

(* to reduce numerical errors we normalize the solution *)
rowsum=Table[0,{size}];
For[i=1,i<=msize,i++,
rowsum+=XX[[i]].hh;
];
XX[[1]]-=DiagonalMatrix[rowsum];

Return[XX];
];



MRAPToMMAP[X_]:= 
Module[{size, msize, \[Epsilon], i},
msize=Dimensions[X][[1]];
size=Dimensions[X[[1]]][[1]];
\[Epsilon]=10^-8;  (* numerical precision *)
If[ Max[Abs[Sum[X[[i]],{i,1,msize}].Table[1,{size}] ]]> \[Epsilon],
Print["Rowsum is different from 0!! (Precision:",\[Epsilon]//N,")"];
Return[X]
];
Return[RationalToMarkovian[X,size, MMAPTransFunc, MMAPEvalFunc, MMAPCheckFunc]];
];



MMAPCheckFunc[X_] :=
Module[{\[Epsilon],c,size,msize,ii,h0},
msize=Dimensions[X][[1]];
size=Dimensions[X[[1]]][[1]];
\[Epsilon]=N[10^-10];
For[ii=2,ii<=msize,ii++,
  If[ Min[ X[[ii]] ] < -\[Epsilon] , Return[False]];
];
h0=X[[1]];
c=-Min[h0];
If[Min[h0+c IdentityMatrix[size]]<-\[Epsilon],Return[False]];
Return[True];
];

MMAPTransFunc[X_,i_,j_,a_] :=
Module[ {bb,size,msize,Y,ii},
msize=Dimensions[X][[1]];
size=Dimensions[X[[1]]][[1]];
Y=Table[0,{msize},{size},{size}];
bb=IdentityMatrix[size];
bb[[i,j]]=a;
bb[[i,i]]=1-a;
For[ii=1,ii<=msize,ii++,
  Y[[ii]]=Inverse[bb].X[[ii]].bb;
];
Return[Y];
];

MMAPEvalFunc[X_,Error_] :=
Module[ {i,j,sum,size,msize,ii},
msize=Dimensions[X][[1]];
size=Dimensions[X[[1]]][[1]];
sum=0;
For[i=1,i<=size,i++,
  For[j=1,j<=size,j++,
    If[ i!=j, sum+=Error[ X[[1]][[i,j]] ] ];
  ];
];
For[ii=2,ii<=msize,ii++,
  For[i=1,i<=size,i++,
    For[j=1,j<=size,j++,
      sum+=Error[ X[[ii]][[i,j]] ];
    ];
  ];
];
Return[sum];
];


TRAPToTMAP[\[Alpha]_,h0_,h1_]:= 
Module[{size},
size=Dimensions[h0][[1]];
X={\[Alpha],h0,h1};
Return[RationalToMarkovian[ X,size, TMAPTransFunc, TMAPEvalFunc, TMAPCheckFunc]];
];


TMAPCheckFunc[X_] :=CheckTMAP[X[[1]],X[[2]],X[[3]]];
CheckTMAP[\[Alpha]_,h0_,h1_]:=
Module[{\[Epsilon],c,size},
\[Epsilon]=N[10^-10];
If[Min[\[Alpha]]<-\[Epsilon],Return[False]];
If[Min[h1]<-\[Epsilon],Return[False]];
size=Dimensions[h0][[1]];
c=-Min[h0];
If[Min[h0+c IdentityMatrix[size]]<-\[Epsilon],Return[False]];
Return[True];
];
TMAPTransFunc[X_,i_,j_,a_] :=
Module[ {bb,size},
size=Dimensions[X[[1]]][[1]];
bb=IdentityMatrix[size];
bb[[i,j]]=a;
bb[[i,i]]=1-a;
Return[{X[[1]].bb,Inverse[bb].X[[2]].bb,Inverse[bb].X[[3]].bb}]
];
TMAPEvalFunc[X_,Error_] :=
Module[ {i,j,sum,size, \[Alpha], h0,h1,a},
size=Dimensions[X[[2]]][[1]];
\[Alpha]=X[[1]];
h0=X[[2]];
h1=X[[3]];
a=-(h0+h1).Table[1,{size}]; (* exit rate *)
sum=0;
For[i=1,i<=size,i++,
sum+=Error[ \[Alpha][[i]]  ];];
For[i=1,i<=size,i++,For[j=1,j<=size,j++,
If[i!=j,sum+=Error[h0[[i,j]]  ] ];];];
For[i=1,i<=size,i++,For[j=1,j<=size,j++,
sum+=Error[h1[[i,j]]  ];];];
For[i=1,i<=size,i++,
sum+=Error[ a[[i]]  ];];
Return[sum];
];



RBTToMBT[\[Alpha]_,h0_,b_]:= 
Module[{size},
size=Dimensions[h0][[1]];
X={\[Alpha],h0,b};
Return[RationalToMarkovian[ X,size, MBTTransFunc, MBTEvalFunc, MBTCheckFunc]];
];



MBTCheckFunc[X_] :=CheckMBT[X[[1]],X[[2]],X[[3]]];
CheckMBT[\[Alpha]_,h0_,b_]:=
Module[{\[Epsilon],c,size},
\[Epsilon]=N[10^-10];
If[Min[\[Alpha]]<-\[Epsilon],Return[False]];
If[Min[b]<-\[Epsilon],Return[False]];
size=Dimensions[h0][[1]];
c=-Min[h0];
If[Min[h0+c IdentityMatrix[size]]<-\[Epsilon],Return[False]];
Return[True];
];
MBTTransFunc[X_,i_,j_,a_] :=
Module[ {bb,size},
size=Dimensions[X[[1]]][[1]];
bb=IdentityMatrix[size];
bb[[i,j]]=a;
bb[[i,i]]=1-a;
Return[{X[[1]].bb,Inverse[bb].X[[2]].bb,Inverse[bb].X[[3]].KroneckerProduct[bb,bb]}]
];
MBTEvalFunc[X_,Error_] :=
Module[ {i,j,sum,size, \[Alpha], h0,b,a},
size=Dimensions[X[[2]]][[1]];
\[Alpha]=X[[1]];
h0=X[[2]];
b=X[[3]];
a= - h0.Table[1,{size}] - b.Table[1,{size^2}]; (* exit rate *)
sum=0;
For[i=1,i<=size,i++,
sum+=Error[\[Alpha][[i]]  ];];
For[i=1,i<=size,i++,For[j=1,j<=size,j++,
If[i!=j,sum+=Error[h0[[i,j]]  ] ];];];
For[i=1,i<=size,i++,
For[j=1,j<=size^2,j++,
sum+=Error[b[[i,j]]  ];];];
For[i=1,i<=size,i++,
sum+=Error[ a[[i]]  ];];
Return[sum];
];


End[(* Private *)]

Protect[RBTObsMinimize, RBTContMinimize, RBTMinimize, GammaijkMomentsFromRBT, MomentsFromRBT, RBTFromMoments, 
TRAPObsMinimize, TRAPContMinimize, TRAPMinimize, MomentsFromTRAP, LagkJointMomentsFromTRAP, TRAPFromMoments,
MStairCase, MRAPObsMinimize, MRAPContMinimize, MRAPMinimize, MomentsFromMRAP, Lag1JointMomentsFromMRAP, MRAPFromMoments, 
MRAPToMMAP, TRAPToTMAP, RBTToMBT ];

EndPackage[]
