(* ::Package:: *)

(*
   BuTools_Utility Package
*)

BeginPackage["BuTools`Utilities`"];

BuToolsVerbose::usage = "BuToolsVerbose is a flag that enables verbose messages. The default value is False."

CTMCSolve::usage = "CTMCSolve[ mx ] -> [ vector ] gives the steady state distribution of the CTMC with generator matrix x."

CRPSolve::usage = "CRPSolve [ mx ] -> [ vector ] gives the steady state distribution of the CRP (without generator matrix check)"

DRPSolve::usage = "DRPSolve [ mx ] -> [ vector ] gives the steady state distribution of the DRP"

DTMCSolve::usage = "DTMCSolve[ mx ] -> [ vector ] gives the steady state distribution of the DTMC with transition probability matrix mx."

MomsToNormMoms::usage = "MomsToNormMoms [ moments ] -> [ moments ] makes normalized moments from moments"

NormMomsToMoms::usage = "NormMomsToMoms [ moments ] -> [ moments ] makes moments from normalized moments"

CheckMAP::usage = "CheckMAP [matrix0, matrix1] checks if the input matrixes define a continuous time MAP (The matrix0 and matrix1 are equal squrare matrixes, matrix0 is a transient generator matrix, matrix1 has only positive or zero elements, and the rowsum of matrix0+matrix1 is 0)"

CheckTransientGenerator::usage = "CheckTransientGenerator [matrix] checks if the matrix is a valid transient generator matrix (the matrix is a square matrix, the diagonal of the matrix is negative, the matrix has positive or zero off-diagonal elements, the real part of the maximum absolute eigenvalue is less than zero) "

CheckGenerator::usage = "CheckGenerator [ matrix ] checks if the matrix is a valid generator matrix (the matrix is a square matrix, the matrix has positive or zero off-diagonal elements, the diagonal of the matrix is negative, the rowsum of the matrix is 0) "

CheckProbMatrix::usage = "CheckProbMatrix [ matrix ] checks if the matrix is a valid probability matrix (the matrix is a square matrix, the matrix has positive or zero off-diagonal elements, the rowsum of the matrix is 1)"

CheckTransientProbMatrix::usage = "CheckTransientProbMatrix [ matrix ] checks if the matrix is a valid transient probability matrix (the matrix is a square matrix, the matrix has positive or zero off-diagonal elements, the rowsum of the matrix is less equal than 1, the maximum absolute eigenvalue is less than 1)"

CheckProbVector::usage = "CheckProbVector [ vector ] checks if the vector is a valid probability vector (the vector has only non-negative elements, the sum of the vector elements is 1)"

CheckSubProbVector::usage = "CheckSubProbVector [ vector ] checks if the vector is a valid sub-probability vector (the vector has only non-negative elements, the sum of the vector elements is less equal than 1)"

MatrixCopy::usage = "MatrixCopy [ matrix0, matrix1, i, j ] -> [ matrix ] gives back the matrix which is obtained from matrix matrix0 by copying matrix1 into it from posiotion [[i, j]] "

KroneckerSum::usage = "KroneckerSum [ matrix0, matrix1 ] -> [matrix ] gives the Kronecker sum of the two matrices"

MomsToReducedmoms::usage = "MomsToReducedmoms [ moments ] -> [ moments ] computes reduced moments from moments"

ReducedmomsToMoms::usage = "ReducedmomsToMoms [ moments ] -> [ moments ] computes moments from reduced moments"

CheckPH::usage = "CheckPH [ vector, matrix ] checks the phase type conditions on a vector and a matrix (the vector is a probability vector, the matrix is a transient generator and they have the same size)"

CheckME::usage = "CheckME [ vector, matrix ] checks some matrix exponential conditions of a vector-matrix pair (the matrix is a square matrix, the vector and the matrix have the same size, the dominant eigenvalue is negative and real)"

CheckRAP::usage = "CheckRAP [ matrix0, matrix1 ] checks the rational arrival process conditions on 2 matrices (matrix0, matrix1 are square matrices, they have the same size, the dominant eigenvalue of matrix0 is negative and real)"

RationalToMarkovian::usage = "RationalToMarkovian -- it is a general procedure to transform non-Markovian representations to Markovian representations. "
ErrorMinimize::usage = "ErrorMinimize -- it is an internal function of RationalToMarkovian. "
ElementaryStep::usage = "ElementaryStep -- it is an internal function of RationalToMarkovian. "

BuToolsReal::usage = "BuToolsReal [ x, n ] sets to zero the imaginary part of x if it is n orders of magnitude smaller than the real part. "



(* ::Input:: *)
(**)


Unprotect[CTMCSolve, DTMCSolve, MomsToNormMoms, 
NormMomsToMoms, CheckMAP, CheckTransientGenerator, CheckGenerator,
 CheckProbMatrix, CheckTransientProbMatrix, CheckProbVector, CheckSubProbVector,
MatrixCopy, KroneckerSum, MomsToReducedmoms,
ReducedmomsToMoms, CheckPH, CheckME, CheckRAP, CRPSolve, DRPSolve, RationalToMarkovian, ErrorMinimize, ElementaryStep,
BuToolsReal];



Begin["`Private`"]


BuToolsVerbose := False 



CTMCSolve[mbe_] :=
Module[ {mx,isize, p0,i},
	If[Not[CheckGenerator[mbe]],Return[{0}]];
	CRPSolve[mbe] 
	]


CRPSolve[mbe_] :=
Module[ {mx,isize, p0,i},
	mx=mbe;
	isize=Dimensions[mx][[1]];
	p0=Table[0,{isize}];
	p0[[1]]=1 ;
	Do[mx[[i,1]]=1,{i, isize}];
	mx= Transpose[mx];
	LinearSolve[mx, p0]
	]



DRPSolve[mbe_] := CRPSolve[mbe-IdentityMatrix[Dimensions[mbe][[1]]]];


DTMCSolve[mbe_]:= 
Module[ {isize, i},
	isize=Dimensions[mbe][[1]];
	i= IdentityMatrix[isize];
  CTMCSolve[mbe-i]
	]


MomsToNormMoms[ moms_] :=
Module[ {isize, i, normmoms ,m},
isize=Dimensions[moms][[1]];
normmoms={moms[[1]]};
For[ i= 2,i<= isize,  m=moms[[i]]/(moms[[i-1]] moms[[1]]);
normmoms=Append[normmoms,m];i++];
normmoms
];


NormMomsToMoms[ normmoms_] :=
Module[ {isize, i, moms ,m},
isize=Dimensions[normmoms][[1]];
moms={normmoms[[1]]};
For[ i= 2,i<= isize,  m=normmoms[[i]] moms[[i-1]] moms[[1]];
moms=Append[moms,m];i++];
moms
];


CheckMAP[h0_,h1_] :=
Module[ {zero,h,size1,size2,\[Epsilon]},
\[Epsilon]=N[10^-10];
If[Not[CheckTransientGenerator[h0]],Return[False]];

{size1,size2}=Dimensions[h0];
If[{size1,size2}!=Dimensions[h1],
If[BuToolsVerbose==True,Print["CheckMAP: matrix0 and matrix1 have different sizes!"]]; 
Return[False]
];

If[ Min[h1]<0,
If[BuToolsVerbose==True,Print["CheckMAP: matrix1 has negative element!"]];
Return[False]
];

h=Table[1,{size1}];
zero=Table[0,{size1}];
If[  Max[Abs[(h0+h1).h]]>\[Epsilon],
If[BuToolsVerbose==True,Print["CheckMAP: A rowsum of matrix0+matrix1 is not 0 (at precision ",\[Epsilon],")!!"]];
Return[False]
];

Return[True]
];


CheckTransientGenerator[h0_]:=
Module[{k,myh0,zero,h,size1,size2,\[Epsilon]},
\[Epsilon]=N[10^-10];
{size1,size2}=Dimensions[h0];
If[size1!=size2,
If[BuToolsVerbose==True,Print["CheckTransientGenerator: the generator is not a square matrix!"]];
Return[False]
];

myh0=h0;
For[k=1,k<=size1,k++,
If[myh0[[k,k]]>=0,
If[BuToolsVerbose==True,
Print["CheckTransientGenerator: The diagonal of the generator is not negative!"]];
Return[False],
myh0[[k,k]]=0];
];

If[Min[myh0]<0,
If[BuToolsVerbose==True,Print["CheckTransientGenerator: matrix0 has negative off-diagonal element!"]];
Return[False]
];

h=Table[1,{size1}];
If[Max[h0.h]>0,
If[BuToolsVerbose==True,Print["CheckTransientGenerator: A rowsum of the generator is greather than 0!"]];
Return[False];
];

If[Max[Re[Eigenvalues[h0//N]]]>=-\[Epsilon],
If[BuToolsVerbose==True,Print["CheckTransientGenerator: the generator has non-negative eigenvalue (at precision ",\[Epsilon],")!"]];
Return[False]
];

Return[True];
];


CheckGenerator[h0_]:=
Module[ {k,myh0,h,size1,size2,\[Epsilon]},
\[Epsilon]=N[10^-10];
{size1,size2}=Dimensions[h0];
If[size1!=size2,
If[BuToolsVerbose==True,Print["CheckGenerator: the generator is not a square matrix!"]];
Return[False]
];

myh0=h0;
For[k=1,k<=size1,k++,
If[myh0[[k,k]]>=0,
If[BuToolsVerbose==True,Print["CheckGenerator: The diagonal of the generator is not negative!"]];
Return[False],
myh0[[k,k]]=0
];
];

If[ Min[myh0]<0,
If[BuToolsVerbose==True,Print["CheckGenerator: the generator has negative off-diagonal element!"]];
Return[False]
];

h=Table[1,{size1}];
If[  Max[Abs[h0.h]]>\[Epsilon], 
If[BuToolsVerbose==True,Print["CheckGenerator: A rowsum of the generator is not 0 (precision:",\[Epsilon],")!!"]];
Return[False]
];

Return[True]
];


CheckProbMatrix[h0_]:=
Module[{size1, size2, h,\[Epsilon]},
\[Epsilon]=N[10^-10];
{size1,size2}=Dimensions[h0];
If[size1!=size2,
If[BuToolsVerbose==True,Print["CheckProbMatrix: the matrix is not a square matrix!"]];
Return[False]
];

If[ Min[h0]<0,
If[BuToolsVerbose==True,Print["CheckProbMatrix: the matrix has negative element!"]];
Return[False]
];

h=Table[1,{size1}];
If[ Max[Abs[h0.h-h]]>\[Epsilon],
If[BuToolsVerbose==True,Print["CheckProbMatrix: A rowsum of the matrix is not 1 (precision:",\[Epsilon],")!!"]]; 
Return[False]
];

Return[True]
];


CheckTransientProbMatrix[h0_]:=
Module[{size1, size2, h, zero},
{size1,size2}=Dimensions[h0];
If[size1!=size2,
If[BuToolsVerbose==True,
Print["CheckTransientProbMatrix: the matrix is not a square matrix!"]]; 
Return[False]
];

If[ Min[h0]<0,
If[BuToolsVerbose==True,
Print["CheckTransientProbMatrix: the matrix has negative element!"]];
Return[False]
];

h=Table[1,{size1}];
zero=Table[0,{size1}];
If[ Not[h0.h<= h],
If[BuToolsVerbose==True,
Print["CheckTransientProbMatrix: A rowsum of the matrix is not less or equal than 1!"]];
Return[False]
];

If[Not[Max[Re[Eigenvalues[h0//N]]]<1],
If[BuToolsVerbose==True,
Print["CheckTransientProbMatrix: The real part of the largest eigenvalue is not less than 1!"]];
Return[False]
];

Return[True]
];


CheckProbVector[h0_]:=
Module[{\[Epsilon]},
\[Epsilon]=N[10^-10];
If[ Min[h0]<0,
If[BuToolsVerbose==True,Print["CheckProbVector: The vector has negative element!"]];
Return[False]
];

If[ Abs[Total[h0]-1]>\[Epsilon],
If[BuToolsVerbose==True,Print["CheckProbVector: The sum of the vector is not 1 (precision:",\[Epsilon],")!"]];
Return[False]
];
Return[True]
];


CheckSubProbVector[h0_]:=
Module[{h, size1},
 If[ Min[h0]<0,
If[BuToolsVerbose==True,
Print["CheckSubProbVector: The vector has negative element!"];
];
Return[False]
];
If[ Not[Total[h0]<1],
If[BuToolsVerbose==True,
Print["CheckSubProbVector: The sum of the vector is not less than 1!"];
];
Return[False]
];
Return[True]
];


MatrixCopy[mbe1_, mbe2_, ifrom_, jfrom_] :=
Module[ {mx,isize,jsize,i,j, size3, size4, size1, size2},
{size1,size2}=Dimensions[mbe1];
{size3,size4}=Dimensions[mbe2];

If[ifrom + size3 > size1+1 ,
If[BuToolsVerbose==True,
Print["MatrixCopy: Can't copy matrix1 to matrix0: matrix0 has not enough coloumns!"]; 
];
Return[False]
];
If[ jfrom + size4 > size2+1 ,
If[BuToolsVerbose==True,
Print["MatrixCopy: Can't copy matrix1 to matrix0: matrix0 has not enough rows!"];
]; 
Return[False]
];

		mx=mbe1;
		{isize,jsize}=Dimensions[mbe2];
		For[i= 1,i <= isize,
			      For[ j= 1,j<= jsize ,
	               mx[[ifrom+i-1,jfrom+j-1]] =mbe2[[i,j]]; 
					 j++]; 
			i++ ] ;	
		mx
];





KroneckerSum[mbe1_, mbe2_] :=
Module[ {isize1,jsize1,isize2,jsize2, size1, size2, size3, size4},
{size1,size2}=Dimensions[mbe1];
{size3,size4}=Dimensions[mbe2];
If[size1!=size2,
If[BuToolsVerbose==True,Print["KroneckerSum: matrix0 is not a square matrix!"]];
 Return[False]
];
If[size3!=size4,
If[BuToolsVerbose==True,Print["KroneckerSum: matrix1 is not a square matrix!"]]; 
 Return[False]
];
{isize1,jsize1}=Dimensions[mbe1];
{isize2,jsize2}=Dimensions[mbe2];
KroneckerProduct[mbe1,IdentityMatrix[isize2]] +
KroneckerProduct[IdentityMatrix[isize1] ,mbe2] 
]		


MomsToReducedmoms[moms_]:=
Table[moms[[i]]/i!,{i,Dimensions[moms][[1]]}];


ReducedmomsToMoms[ rmoms_] :=
Table[rmoms[[i]]  i!,{i,Dimensions[rmoms][[1]]}];


CheckPH[vector_, h0_]:=
Module[{size1, size2},

If[ Dimensions[vector][[1]]!=Dimensions[h0][[1]],
If[BuToolsVerbose==True,Print["CheckPH: the vector and the matrix have different sizes!"]];
Return[False]
];

If[Not[CheckProbVector[vector]],Return[False]];
If[Not[CheckTransientGenerator[h0]],Return[False]];

Return[True]
];


CheckME[vector_, h0_]:=
Module[{size,eig,ceig,reig,\[Epsilon]},
\[Epsilon]=N[10^-12];

If[ Dimensions[h0][[1]]!=Dimensions[h0][[2]],
If[BuToolsVerbose==True,Print["CheckME: The matrix is not a quadratic matrix!"]];
Return[False]
];

If[ Dimensions[vector][[1]]!=Dimensions[h0][[1]],
If[BuToolsVerbose==True,Print["CheckME: The vector and the matrix have different sizes!"]];
Return[False]
];

eig=Eigenvalues[h0//N];
If[Max[Re[eig]]>=0,
If[BuToolsVerbose==True,Print["CheckME: There is an eigenvalue of the matrix with not negative real part!"]];
 Return[False]
];

size=Dimensions[h0][[1]];
ceig=Table[Switch[Element[eig[[i]],Reals],True,-Infinity,False,Re[eig[[i]]]],{i,size}];
reig=Table[Switch[Element[eig[[i]],Reals],True,Re[eig[[i]]],False,-Infinity],{i,size}];

If[Max[reig]+\[Epsilon]<Max[ceig],
If[BuToolsVerbose==True,Print["CheckME: The dominant eigenvalue of the matrix is not real!"]];
 Return[False]
];

If[Abs[Max[reig]-Max[ceig]]<\[Epsilon],
Print["CheckME: The dominant and a complex eigenvalue has the same real part !!! "]
];

Return[True] ;
];


CheckRAP[h0_,h1_] :=
Module[ {h,size1,size2,size,eig,ceig,reig,\[Epsilon]},
\[Epsilon]=N[10^-10];

If[ Dimensions[h0][[1]]!=Dimensions[h0][[2]],
If[BuToolsVerbose==True,Print["CheckRAP: matrix0 is not a quadratic matrix!"]];
Return[False]
];

If[ Dimensions[h1][[1]]!=Dimensions[h1][[2]],
If[BuToolsVerbose==True,Print["CheckRAP: matrix1 is not a quadratic matrix!"]];
Return[False]
];

{size1,size2}=Dimensions[h0];
If[{size1,size2}!=Dimensions[h1],
If[BuToolsVerbose==True,Print["CheckRAP: matrix0 and matrix1 have different sizes!"]]; 
Return[False]
];

h=Table[1,{size1}];
If[ Max[Abs[(h0+h1).h]] > \[Epsilon],
If[BuToolsVerbose==True,Print["CheckRAP: A rowsum of matrix0+matrix1 is not 0! (precision:",\[Epsilon],")"]];
Return[False]
];

eig=Eigenvalues[h0//N];
If[Max[Re[eig]]>=0,
If[BuToolsVerbose==True,Print["CheckRAP: there is an eigenvalue of matrix0 with non-negative real part!"]];
Return[False]
];

size=Dimensions[h0][[1]];
ceig=Table[Switch[Element[eig[[i]],Reals],True,-Infinity,False,Re[eig[[i]]]],{i,size}];
reig=Table[Switch[Element[eig[[i]],Reals],True,Re[eig[[i]]],False,-Infinity],{i,size}];

If[Max[reig]<Max[ceig],
If[BuToolsVerbose==True,Print["CheckRAP: The dominant eigenvalue of matrix0 is not real!"]];
 Return[False]
];

If[Max[reig]==Max[ceig],
Print["CheckRAP: The dominant and a complex eigenvalue of matrix0 has the same real part !!! "]
];

Return[True]
];


RationalToMarkovian[Xin_, size_,TransFunc_,EvalFunc_,CheckFunc_] :=
Module[ {gsum,sum, n, X, Y,  nstepsize,nrepeat,niter,nerrorfunc,errortype,stepsize,i,j,ErrorFunc,x},
X=Xin;
If[CheckFunc[X],
Print["RationalToMarkovian is called with a Markovian representation !!!"];
Return[X];
];
gsum=0;
nstepsize=7;(*10;*) (* a-t csokkentem 2^-1 tol 2^-nstepsize-ig *)
nrepeat=1;(*2;*) (* minden a-ra n2 kort fut *)
niter=50; (*500;*) (* adott celfuggvenyre max niter iteraciot csinal *)
nerrorfunc=4; (* celfuggvenyre max n3 iteraciot csinal *)

stepsize=1.;
For[i=1,i<=nstepsize,i++,
stepsize=stepsize/2;
For[j=1,j<=nrepeat,j++,
sum=0;
For[errortype=1,errortype<=nerrorfunc,errortype++,
If[errortype==1,ErrorFunc[x_]:=E^(-100 x)];
If[errortype==2,ErrorFunc[x_]:=E^(-1000 x)];
If[errortype==3,ErrorFunc[x_]:=E^(-2 x)];
If[errortype==4,ErrorFunc[x_]:=E^-(x-1)^3];

{X,n}=ErrorMinimize[X,size,niter,stepsize,ErrorFunc,TransFunc,EvalFunc,CheckFunc];
If[CheckFunc[X],
Print["Found a Markovian representation in ", gsum+sum+n, " iterations !!!"];
Return[X];
];
sum+=n;
];
];
gsum+=sum;
];
Print["No MARKOVIAN representation is found in ", gsum, " iterations !!!"];
Return[X];
];


ErrorMinimize[Xin_,size_,maxiter_,stepsize_,ErrorFunc_,TransFunc_,EvalFunc_,CheckFunc_] :=
Module[ {bestX, X, lastdist,improve,i},
X=Xin;
For[i=1,i<=maxiter,i++,
{X,improve}=ElementaryStep[X,size,stepsize,ErrorFunc,TransFunc,EvalFunc,CheckFunc];

If[ Not[ improve] ,Return[{X,i}]];
];
Return[{X,i-1}];
];


ElementaryStep[Xin_,size_,stepsize_,ErrorFunc_,TransFunc_,EvalFunc_,CheckFunc_] :=
Module[ {X, Xu, m,min,mini,minj,mina,i,j},
X=Xin;
min=EvalFunc[X,ErrorFunc];
mini=0;minj=0;
For[i=1,i<=size,i++,
For[j=1,j<=size,j++,
If[i!=j,
Xu=TransFunc[X,i,j,stepsize];
If[ CheckFunc[Xu],Print["MARKOVIAN REPRESEANTAION !!!!!"];Return[{Xu,False}];];
m=EvalFunc[Xu,ErrorFunc];
If[m <min,mini=i;minj=j;min=m ;mina=stepsize];

Xu=TransFunc[X,i,j,-stepsize];
If[ CheckFunc[Xu],Print["MARKOVIAN REPRESEANTAION !!!!!"];Return[{Xu,False}];];
m=EvalFunc[Xu,ErrorFunc];
If[m <min,mini=i;minj=j;min=m ;mina=-stepsize];
];
];
];
If[mini!=0,
Xu=TransFunc[X,mini,minj,mina];
Return[{Xu,True}],
Return[{X,False}];
];
];


BuToolsReal[x_,n_]:=If[Im[x]==0||Abs[Re[x]/Im[x]]>10^n,Return[Re[x]],Return[x]];


End[(* Private *)]

Protect[CTMCSolve, DTMCSolve, MomsToNormMoms, NormMomsToMoms,
 CheckMAP, CheckTransientGenerator, CheckGenerator, CheckProbMatrix,
 CheckTransientProbMatrix, CheckProbVector, CheckSubProbVector,
MatrixCopy, KroneckerSum, MomsToReducedmoms,
ReducedmomsToMoms, CheckPH, CheckME, CheckRAP, CRPSolve, DRPSolve, RationalToMarkovian, ErrorMinimize, ElementaryStep,
BuToolsReal];

EndPackage[]
